/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 bbftp.c  v 0.0.0  1999/11/24
          v 1.2.0  2000/01/24
          v 1.3.0  2000/03/16	- adapt ndport to filelen 
          v 1.4.0  2000/03/22	- use new encryption protocol
		  v 1.5.0  2000/03/24	- Complete change in the loop on the
		  						  input file in order to implement the retry
								- Add a result file more simple in order to
								  help to see which transfer has failed
		  v 1.6.0  2000/03/26	- Add version print in client
		  						- Add code for getting a file
		  v 1.6.1  2000/03/30	- In case of broken connection on the last try
		  						  indicate for the next transfert to retry 
								  connection
		  v 1.7.0  2000/03/30	- Implement getrfio
		  v 1.7.1  2000/04/04	- Portage to SGI (IRIX64)
		  v 1.8.3  2000/04/18	- Add a return code to all transfer function
		  						  (-2 added)
		  v 1.8.5  2000/04/28	- Add makeadir call
		  v 1.8.7  2000/05/28	- Portage to OSF1
		  						- Add starting date
								- Change headers
		  v 1.8.8  2000/06/02	- Add cd command
		  						- Add function reconnecttoserver to setup the
								  correct directory in case of broken connection
		  						- Add lcd command
								- Add mput command
								- Add mget command
		  v 1.8.9  2000/08/08	- Add timestamp
		  v 1.8.10 2000/08/08	- Portage to Linux
		  v 1.8.12 2000/08/14	- Add putrfio command
		  v 1.9.0  2000/08/21	- Use configure to help portage
								- Add -r option

*****************************************************************************/
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <bbftp.h>
#include <client.h>
#include <common.h>
#include <structures.h>
#include <version.h>

#define OPTIONS	"bcdi:o:p:r:tu:v"

struct	sockaddr_in data_source[MAXPORT] ;
int		datasock[MAXPORT] ; 
int		pid_child[MAXPORT] ;
int		debug = 0 ;
int		timestamp = 0 ;
FILE	*infd;
struct sockaddr_in hisctladdr ;
struct sockaddr_in myctladdr ;
char	*remotedir = NULL ;
char	*localdir = NULL ;
pid_t	fatherpid ;
int		unlinkfile ;
char	localfile[MAXLINE] ;
int		globaltrymax = NBTRYMAX ;

void Usage() 
{
	printf("Usage bbftp -i inputfile\n") ;
	printf("            -u username\n") ;
	printf("           [-o outputfile]\n") ;
	printf("           [-p number of // streams]\n") ;
	printf("           [-b (background)]\n") ;
	printf("           [-c (compress)]\n") ;
	printf("           [-t (timestamp)]\n") ;
	printf("           [-r number of tries ]\n") ;
	printf("           host\n") ;
	printf("      or \n") ;
	printf("      bbftp -v\n") ;
	_exit(1) ;  
}
main(argc, argv, envp)
	int argc;
	char **argv;
	char **envp;
{
	extern char *optarg;
	extern int optind, opterr, optopt;
	int		j, retcode, i ;
	char	*inputfile = NULL ;
	char	*outputfile = NULL ;
	char	*username = NULL ;
	char	*password = NULL ;
	char	*hostname ;
	int		hosttype ;
	char	*calchostname ;
	int		nbport = 1 ;
	struct hostent  *hp = NULL ;
	int		stderrfd ;
	int		controlsock ;
	int		background = 0 ;
	int		compress = NOCOMPRESSION ;

	int		nbtry ;
	int		connectionisbroken ;
	
	char	resultfile[MAXLINE] ;
	FILE	*resfd ;

	char	line[MAXLINE] ;
	int		nbitem ;
	char	action[MAXLINE] ;
	char	reste[MAXLINE] ;
	
	time_t  logtime ;
	char	clogtime[50] ;

	if ( blockallsignals() < 0 ) {
		printf("Error in blockallsignals function\n") ;
		_exit(1) ;
	}
	opterr = 1 ;
	while ((j = getopt(argc, argv, OPTIONS)) != -1) {
		switch (j) {
			case 'b' :{
				background = 1 ;
				break ;
			}
			case 'c' :{
				compress = COMPRESSION ;
				break ;
			}
			case 'd' :{
				debug = 1 ;
				break ;
			}
			case 'i' :{
				inputfile = optarg ;
				if ( strlen(inputfile)+4 > MAXLINE ) {
					fprintf(stderr,"Input file name too long %s\n",inputfile) ;
					_exit(1) ;
				}
				strcpy(resultfile,inputfile) ;
				strcat(resultfile,".res") ;
				break ;
			}
			case 'o' :{
				outputfile = optarg ;
				break ;
			}
			case 'p' :{
				retcode = sscanf(optarg,"%d",&nbport) ;
				if ( retcode != 1 ) {
					fprintf(stderr,"Number of port must be numeric\n") ;
					_exit(1) ;
				}
				if ( nbport <= 0 || nbport > MAXPORT ) {
					fprintf(stderr,"Number of port must be less than %d\n",MAXPORT) ;
					_exit(1) ;
				}
				break ;
			}
			case 'r' :{
				retcode = sscanf(optarg,"%d",&globaltrymax) ;
				if ( retcode != 1 ) {
					fprintf(stderr,"Number of tries must be numeric\n") ;
					_exit(1) ;
				}
				break ;
			}
			case 't' :{
				timestamp = 1 ;
				break ;
			}
			case 'u' :{
				username = optarg ;
				break ;
			}
			case 'v' :{
				printf("bbftp version %s compiled with %d tries\n",VERSION,NBTRYMAX) ;
				exit(1) ;
			}
			default : {
				Usage() ;
			}
		}
	}
	if ( optind != argc-1 ) Usage() ;
	hostname = argv[optind] ;
	hosttype = 0 ;

	for (i=0 ; i < strlen(hostname) ; i++) {
		if ( isalpha(hostname[i]) ) {
			hosttype = 1 ;
			break ;
		} else if ( isdigit(hostname[i]) ) {
		} else if ( hostname[i] == '.' ) {
		} else {
			printf("Invalid hostname : %s\n",hostname) ;
			_exit(1) ;
		}
	}
	if ( hosttype == 0 ) {
		/*
		** Adresse numerique
		*/
		hisctladdr.sin_addr.s_addr = 0 ;
		hisctladdr.sin_addr.s_addr = inet_addr(hostname) ;
		if (hisctladdr.sin_addr.s_addr == -1 ) {
			fprintf(stderr,"Incorrect address inet_addr %s\n",hostname) ;
			_exit(1) ;
		}
		calchostname = (char *)inet_ntoa(hisctladdr.sin_addr) ;
		if ( strcmp(hostname,calchostname) ) {
			fprintf(stderr,"Incorrect address %s\n",hostname) ;
			_exit(1) ;
		}
	} else {
		if ( (hp = gethostbyname(hostname) ) == NULL ) {
			/*
			*/
			printf("Host %s not found\n",hostname) ;
			_exit(1) ;
		} else {
			/*
			** C'est un nom de machine
 			*/
			if (hp->h_length > (int)sizeof(hisctladdr.sin_addr)) {
				hp->h_length = sizeof(hisctladdr.sin_addr);
			}
			memcpy(&hisctladdr.sin_addr, hp->h_addr_list[0], hp->h_length) ;
		}
	}
	if ( inputfile == NULL ) {
		fprintf(stderr,"No input file \n") ;
		Usage() ;
	}
	if ( username == NULL ) {
		fprintf(stderr,"No username \n") ;
		Usage() ;
	}
/*
** Ask for password
*/
#ifdef USE_GETPASSPHRASE
	password = (char *) getpassphrase("Password: ") ; 
#else
	sprintf(line,"Password (max length = %d): ",sysconf(_SC_PASS_MAX)) ;
	password = (char *) getpass(line) ; 
#endif
	if ( strlen(password) == 0 ) {
		printf("You must enter a password\n") ;
		_exit(1) ;
	}
	
	if ( (infd = fopen(inputfile,"r") ) == NULL) {
		fprintf(stderr,"Unable to open input file %s\n",inputfile) ;
		_exit(1) ;
	}
	if ( (resfd = fopen(resultfile,"w") ) == NULL) {
		fprintf(stderr,"Unable to open result file %s\n",resultfile) ;
		_exit(1) ;
	}
	if ( time(&logtime) == -1 ) {
		strcpy(clogtime,"") ;
	} else {
		if ( strftime(clogtime,sizeof(clogtime),"%a %b %d %H:%M:%S",localtime(&logtime)) == 0 )	{
			strcpy(clogtime,"") ;
		}
	}
	if ( background == 1 ) {
		retcode = fork() ;
		if ( retcode < 0 ) printf("Error forking\n") ;
		if (retcode > 0 ) exit(0) ;
		setsid() ;
		printf("Starting under pid %d on %s\n",getpid(),clogtime)		;
	}
	if ( outputfile != NULL ) {
		stderrfd = open(outputfile,O_CREAT|O_WRONLY|O_SYNC|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
		if ( stderrfd < 0 ) {
			printf("Unable to open output file : Error%d(%s)\n",errno,strerror(errno)) ;
			_exit(1) ;
		}
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
		if ( fcntl(stderrfd,F_DUPFD,STDOUT_FILENO) != STDOUT_FILENO ) {
			printf("Unable to fcntl stdout\n") ;
			_exit(1) ;
		}
		if ( fcntl(stderrfd,F_DUPFD,STDERR_FILENO) != STDERR_FILENO ) {
			printf("Unable to fcntl stderr\n") ;
			_exit(1) ;
		}
	}	
	printf("Starting under pid %d on %s\n",getpid(),clogtime)		;
	fatherpid = getpid() ;
	unlinkfile = 0 ;
	/*
	** Initialize the pid array (for the case of SIGTERM)
	*/
	for ( i=0 ; i< MAXPORT ; i++) {
		pid_child[i] = 0 ;
	}
	if ( setsignals_b() < 0 ) {
		printf("Error in setsignals function\n") ;
		_exit(1) ;
	}

/*
** Connect to the remote host on control port
*/
	for ( nbtry = 1 ; nbtry <= globaltrymax ; nbtry++ ) {
		retcode = connecttoserver(username,password,&controlsock) ;
		if ( retcode < -1 ) {
			_exit(1) ;
		} else if ( retcode == 0 ) {
			break ;
		} else {
			if ( nbtry != globaltrymax ) {
				if (timestamp ) printtime() ;
				printf("Retrying connection waiting %d s\n",WAITRETRYTIME) ;
				sleep(WAITRETRYTIME) ;
			}
		}
	}
	if ( retcode != 0 && nbtry == globaltrymax+1) {
		if (timestamp ) printtime() ;
		printf("Maximum try reached (%d) aborting connection\n",globaltrymax) ;
		_exit(1) ;
	}
	connectionisbroken = 0 ;
/*
** Now we are going to loop on the command in the input file
*/
	while ( fgets(line,MAXLINE,infd) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		nbitem = sscanf(line,"%s %s",action,reste) ;
		if (nbitem == EOF || nbitem == 0 ) {
			if (timestamp ) printtime() ;
			printf("Error in input file %s on line : %s\n",inputfile,line) ;
			fprintf(resfd,"%s FAILED\n",line) ;
			fflush(resfd) ;
		} else {
			/*
			** Decode the action
			*/
			if ( !strcmp(action,"cd") || 
				 !strcmp(action,"get") || 
				 !strcmp(action,"getrfio") ||
				 !strcmp(action,"lcd") ||
				 !strcmp(action,"mget")  ||
				 !strcmp(action,"mkdir")  ||
				 !strcmp(action,"mput")  ||
				 !strcmp(action,"put") ||
				 !strcmp(action,"putrfio") ) {
				for ( nbtry = 1 ; nbtry <= globaltrymax ; nbtry++ ) {
					/*
					** check if the last try was not disconnected
					*/
					if ( connectionisbroken == 1 ) {
						reconnecttoserver(username,password,&controlsock) ;
						connectionisbroken = 0 ;
					}
					/*
					** Initialize the pid array (for the case of SIGTERM)
					*/
					for ( i=0 ; i< MAXPORT ; i++) {
						pid_child[i] = 0 ;
					}
					if (!strcmp(action,"cd") ) {
						retcode = changedir(line,controlsock,nbport,compress) ;
					} else if (!strcmp(action,"get") ) {
						unlinkfile = 0 ;
						retcode = getafile(line,controlsock,nbport,compress,0) ;
						unlinkfile = 0 ;
					} else if (!strcmp(action,"getrfio") ) {
						retcode = getafile(line,controlsock,nbport,compress,1) ;
					} else if (!strcmp(action,"lcd") ) {
						retcode = localchangedir(line) ;
					} else if (!strcmp(action,"mget") ) {
						retcode = mgetfile(line,&controlsock,nbport,compress,resfd,username,password) ;
					} else if (!strcmp(action,"mkdir") ) {
						retcode = makeadir(line,controlsock) ;
					} else if (!strcmp(action,"mput") ) {
						retcode = mputfile(line,&controlsock,nbport,compress,resfd,username,password) ;
					} else if (!strcmp(action,"put") ) {
						retcode = putafile(line,controlsock,nbport,compress) ;
					} else if (!strcmp(action,"putrfio") ) {
						retcode = putafilerfio(line,controlsock,nbport,compress) ;
					}
					if ( retcode == BB_RET_FT_NR ) {
						/*
						** Fatal error no retry and connection is not broken
						*/
						break ;
					} else if ( retcode == BB_RET_FT_NR_CONN_BROKEN ) {
						/*
						** Fatal error no retry and connection is broken
						*/
						connectionisbroken = 1 ;
						break ;
					} else if ( retcode == BB_RET_OK ) {
						/*
						** Success , no retry
						*/							
						break ;
					} else if ( retcode == BB_RET_CONN_BROKEN ) {
						/*
						** Retry needed with a new connection
						*/
						if ( nbtry != globaltrymax ) {
							if (timestamp ) printtime() ;
							printf("Retrying connection waiting %d s\n",WAITRETRYTIME) ;
							sleep(WAITRETRYTIME) ;
							reconnecttoserver(username,password,&controlsock) ;
						} else {
							/*
							** In case of broken connection and max retry reached
							** indicate to the next transfer that the connection
							** is broken
							*/
							connectionisbroken = 1 ;
						}
					} else {
						/*
						** retcode > 0 means retry on this transfer
						*/
						if ( nbtry != globaltrymax ) {
							if (timestamp ) printtime() ;
							printf("Retrying transfer waiting %d s\n",WAITRETRYTIME) ;
							sleep(WAITRETRYTIME) ;
						}
					}
				}
				if ( retcode != 0 && nbtry == globaltrymax+1) {
					if (timestamp ) printtime() ;
					printf("Maximum try reached (%d) skipping file\n",globaltrymax) ;
					fprintf(resfd,"%s FAILED\n",line) ;
					fflush(resfd) ;
				} else if ( retcode == 0 ) {
					fprintf(resfd,"%s OK\n",line) ;
					fflush(resfd) ;
				} else {
					fprintf(resfd,"%s FAILED\n",line) ;
					fflush(resfd) ;
				}
			} else {
				if (timestamp ) printtime() ;
				printf("Unknown command : %s\n",line) ;
				fprintf(resfd,"%s FAILED\n",line) ;
			}
		}
	}
}
#ifndef WORDS_BIGENDIAN
my64_t convertlong(my64_t v) {
	struct bb {
		int	fb ;
		int sb ;
	} ;
	struct bb *bbpt ;
	int 	tmp ;
	my64_t	tmp64 ;
	
	tmp64 = v ;
	bbpt = (struct bb *) &tmp64 ;
	tmp = bbpt->fb ;
	bbpt->fb = ntohl(bbpt->sb) ;
	bbpt->sb = ntohl(tmp) ;	
	return tmp64 ;
}
#endif 
