/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 bbftpc.c  v 1.8.9  2000/08/01	- Program creation
		   v 1.8.10 2000/08/11	- Portage to Linux
		   v 1.8.12 2000/08/14	- Add putrfio command
		   v 1.9.0  2000/08/14	- Add -r option

*****************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/un.h>
#include <unistd.h>


#include <bbftp.h>
#include <client.h>
#include <common.h>
#include <structures.h>
#include <version.h>

#define OPTIONS	"cdbe:i:m:o:p:r:tu:v"

/*
** Specific to c
*/
int		flagsigusr1 ;
int		flagsigusr2 ;
int		flagsighup ;
pid_t	ownserverpid = 0 ;

void Usage() 
{
	printf("Usage bbftpc -i inputfile | -e \"command line\"\n") ;
	printf("             -m magic number\n") ;
	printf("            [-o outputfile]\n") ;
	printf("            [-p number of // streams]\n") ;
	printf("            [-b (background)]\n") ;
	printf("            [-c (compress)]\n") ;
	printf("            [-t (timestamp)]\n") ;
	printf("            [-r number of tries ]\n") ;
	printf("      or \n") ;
	printf("      bbftpc -v\n") ;
	exit(-1) ;  
}
main(argc, argv, envp)
	int argc;
	char **argv;
	char **envp;
{
#ifdef HAVE_SIGINFO
	extern char *optarg;
	extern int optind, opterr, optopt;
	int		j, retcode ;
	char	*inputfile = NULL ;
	FILE	*infd ;
	char	*outputfile = NULL ;
	char	pwddir[PATH_MAX] ;
	char	*commandline = NULL ;

	int		nbport = 1 ;
	int		background = 0 ;
	int		compress = NOCOMPRESSION ;
	int		debug=0 ;
	int		timestamp=0 ;
	

	struct	sockaddr_un unixsocaddr ;
	int		unixsocaddrlen ;
	int		newstdin,newstdout ;
	int		nfds ;
	fd_set	selectmask ; /* Select mask */
	struct timeval	wait_timer;
	pid_t	serverpid = 0 ;

	int		stderrfd ;
	
	char	resultfile[PATH_MAX] ;
	FILE	*resfd ;

	char	line[MAXLINE] ;
	char	lineread[MAXLINE] ;
	int		nbitem ;
	char	action[MAXLINE] ;
	char	reste[MAXLINE] ;
	
	FILE	*toserver, *fromserver ;
	
	time_t  logtime ;
	char	clogtime[50] ;
	int		globaltrymax = NBTRYMAX;
	
	background = 0 ;
	if ( blockallsignals() < 0 ) {
		printf("Error in blockallsignals function\n") ;
		exit(-1) ;
	}

	opterr = 1 ;
	while ((j = getopt(argc, argv, OPTIONS)) != -1) {
		switch (j) {
			case 'b' :{
				background = 1 ;
				break ;
			}
			case 'c' :{
				compress = COMPRESSION ;
				break ;
			}
			case 'd' :{
				debug = 1 ;
				break ;
			}
			case 'e' :{
				commandline =  optarg ;
				break ;
			}
			case 'i' :{
				inputfile = optarg ;
				if ( strlen(inputfile)+4 > PATH_MAX ) {
					printf("Input file name too long %s\n",inputfile) ;
					exit(-1) ;
				}
				strcpy(resultfile,inputfile) ;
				strcat(resultfile,".res") ;
				break ;
			}
			case 'm' :{
				retcode = sscanf(optarg,"%d",&serverpid) ;
				if ( retcode != 1 ) {
					printf("Magic number must be numeric\n") ;
					exit(-1) ;
				}
				break ;
			}
			case 'o' :{
				outputfile = optarg ;
				break ;
			}
			case 'p' :{
				retcode = sscanf(optarg,"%d",&nbport) ;
				if ( retcode != 1 ) {
					printf("Number of port must be numeric\n") ;
					exit(-1) ;
				}
				if ( nbport <= 0 || nbport > MAXPORT ) {
					printf("Number of port must be less than %d\n",MAXPORT) ;
					exit(-1) ;
				}
				break ;
			}
			case 'r' :{
				retcode = sscanf(optarg,"%d",&globaltrymax) ;
				if ( retcode != 1 ) {
					fprintf(stderr,"Number of tries must be numeric\n") ;
					_exit(1) ;
				}
				break ;
			}
			case 't' :{
				timestamp = 1 ;
				break ;
			}
			case 'v' :{
				printf("bbftpc version %s compiled with %d tries\n",VERSION,NBTRYMAX) ;
				exit(0) ;
			}
			default : {
				Usage() ;
			}
		}
	}
	if ( argc == 1 ) Usage() ;
	/*
	** Get the current directory
	*/
	if ( getcwd (pwddir,PATH_MAX) == NULL ) {
		printf("Unable to get current directory : %s \n",strerror(errno)) ;
		exit(-1) ;
	}
	/*
	** Check if input file or commandline (mutualy exclusive)
	*/
	if ( inputfile == NULL ) {
		if ( commandline == NULL ) {
			printf("Nor input file neither expression\n") ;
			Usage() ;
		}
	} else if ( commandline != NULL ) {
		printf("-i and -e options are mutualy exclusive\n") ;
		Usage() ;
	} else {
		if ( (infd = fopen(inputfile,"r") ) == NULL) {
			printf("Unable to open input file %s\n",inputfile) ;
			exit(-1) ;
		}
		if ( (resfd = fopen(resultfile,"w") ) == NULL) {
			printf("Unable to open result file %s\n",resultfile) ;
			exit(-1) ;
		}
	}
	if ( serverpid == 0 ) {
		printf("No Magic number \n") ;
		exit(-1) ;
	}
	
/*
** Get start time
*/
	if ( time(&logtime) == -1 ) {
		strcpy(clogtime,"") ;
	} else {
		if ( strftime(clogtime,sizeof(clogtime),"%a %b %d %H:%M:%S",localtime(&logtime)) == 0 )	{
			strcpy(clogtime,"") ;
		}
	}
/*
** Fork if in background
*/
	if ( background == 1 ) {
		retcode = fork() ;
		if ( retcode < 0 ) printf("Error forking\n") ;
		if ( retcode > 0 ) exit(0) ;
		setsid() ;
		printf("Starting under pid %d on %s\n",getpid(),clogtime)		;
	}
/*
** Change stdout if requested
*/
	if ( outputfile != NULL ) {
		stderrfd = open(outputfile,O_CREAT|O_WRONLY|O_SYNC|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
		if ( stderrfd < 0 ) {
			printf("Unable to open output file : Error%d(%s)\n",errno,strerror(errno)) ;
			exit(-1) ;
		}
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
		if ( fcntl(stderrfd,F_DUPFD,STDOUT_FILENO) != STDOUT_FILENO ) {
			printf("Unable to fcntl stdout\n") ;
			exit(-1) ;
		}
		if ( fcntl(stderrfd,F_DUPFD,STDERR_FILENO) != STDERR_FILENO ) {
			printf("Unable to fcntl stderr\n") ;
			exit(-1) ;
		}
	}
	printf("Starting under pid %d on %s\n",getpid(),clogtime) ;
/*
** Paly with signals
*/
	if ( setsignals_c() < 0 ) {
		printf("Error in setsignals_c function\n") ;
		exit(-1) ;
	}
/*
** Send a USR1 signal to server to tell it is requested
*/
	flagsigusr1 = 2 ;
	kill(serverpid,STARTNEW) ;
/*
** Create the unix socket 
*/
	if ( (newstdin = socket(AF_UNIX,SOCK_STREAM,0) ) < 0 ) {
		printf("Cannot create newstdin socket : %s\n",strerror(errno)) ;
		exit(-1) ;
	}
	if ( (newstdout = socket(AF_UNIX,SOCK_STREAM,0) ) < 0 ) {
		printf("Cannot create newstdout socket : %s\n",strerror(errno)) ;
		exit(-1) ;
	}
/*
** Wait for the USR1 signal telling the unix socket are
** created
*/
	if ( flagsigusr1 == 2 ) {
		nfds = sysconf(_SC_OPEN_MAX) ;
		FD_ZERO(&selectmask) ;
		wait_timer.tv_sec  = 10 ;
		wait_timer.tv_usec = 0 ;
		if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
			/*
			** We receive the signal
			*/
			if ( flagsigusr1 == 0 ) {
				printf("Error while waiting USR1 signal\n") ;
				exit(-1) ;
			}
		} else if ( retcode == 0 ) {
			/*
			** Time out
			*/
			printf("Time out waiting for USR1 server signal\n") ;
			exit(-1) ;
		} else {
		}
	}

/*
** Now connect to the Unix socket
*/
	unixsocaddr.sun_family = AF_UNIX ;
	sprintf(unixsocaddr.sun_path,"/tmp/.bbftpcd.in.%d",getpid()) ;
	unixsocaddrlen = sizeof(unixsocaddr) ;
	
	if ( connect(newstdin,(struct sockaddr*)&unixsocaddr,unixsocaddrlen) < 0 ) {
		printf("Error connect in\n") ;
		exit(-1) ;
	}
	toserver = fdopen(newstdin, "w");	
	unixsocaddr.sun_family = AF_UNIX ;
	sprintf(unixsocaddr.sun_path,"/tmp/.bbftpcd.out.%d",getpid()) ;
	unixsocaddrlen = sizeof(unixsocaddr) ;
	if ( connect(newstdout,(struct sockaddr*)&unixsocaddr,unixsocaddrlen) < 0 ) {
		printf("Error connect out\n") ;
		exit(-1) ;
	}
	fromserver = fdopen(newstdout, "r");	
/*
** Send info
*/
	fprintf(toserver,"%s\n",pwddir) ;
	if ( compress == COMPRESSION ) {
		fprintf(toserver,"on\n") ;
	} else {
		fprintf(toserver,"off\n") ;
	}
	fprintf(toserver,"%d\n",nbport) ;
	if ( debug == 1 ) {
		fprintf(toserver,"on\n") ;
	} else {
		fprintf(toserver,"off\n") ;
	}
	if ( timestamp == 1 ) {
		fprintf(toserver,"on\n") ;
	} else {
		fprintf(toserver,"off\n") ;
	}
	fprintf(toserver,"%d\n",globaltrymax) ;
	fflush(toserver) ;
/*
** If command line just send it and wait for the answer
*/
	if ( commandline != NULL ) {
		nbitem = sscanf(commandline,"%s %s",action,reste) ;
		if (nbitem == EOF || nbitem == 0 ) {
			printf("Error on expression : %s\n",commandline) ;
		} else {
			/*
			** Correct expression
			*/
			if ( !strcmp(action,"cd") || 
				 !strcmp(action,"get") || 
				 !strcmp(action,"getrfio") ||
				 !strcmp(action,"lcd") ||
				 !strcmp(action,"mget")  ||
				 !strcmp(action,"mkdir")  ||
				 !strcmp(action,"mput")  ||
				 !strcmp(action,"put") ||
				 !strcmp(action,"putrfio") ) {
				flagsigusr1 = 0 ;
				flagsigusr2 = 0 ;
				flagsighup = 0 ;
				fprintf(toserver,"%s\n",commandline)  ;
				fflush(toserver) ;
				while ( 1 == 1 ) {
					nfds = sysconf(_SC_OPEN_MAX) ;
					FD_ZERO(&selectmask) ;
					FD_SET(newstdout,&selectmask) ;
					if ( (retcode = select(nfds,&selectmask,0,0,0) ) == -1 ) {
						/*
						** If we got a signal we are here
						*/
						if (flagsigusr2 == 1 ) {
							exit(-1) ;
							break ;
						} else if ( flagsighup == 1 ) {
							exit(0) ;
							break ;
						}
					} else if ( retcode == 0 ) {

					} else {
						if ( FD_ISSET(newstdout,&selectmask) ) {
							if ( fgets(lineread,MAXLINE,fromserver) != NULL ) {
					 			printf("%s",lineread) ;
							}
						}
						if (flagsigusr2 == 1 ) {
							exit(-1) ;
							break ;
						} else if ( flagsighup == 1 ) {
							exit(0) ;
							break ;
						}
					}
				}
			} else {
				printf("Unknown command : %s\n",commandline) ;
			}
		}
	} else {
		/*
		** Loop on the input file
		*/
		while ( fgets(line,MAXLINE,infd) != NULL ) {
			if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
			nbitem = sscanf(line,"%s %s",action,reste) ;
			if (nbitem == EOF || nbitem == 0 ) {
				printf("Error in input file %s on line : %s\n",inputfile,line) ;
				fprintf(resfd,"%s FAILED\n",line) ;
				fflush(resfd) ;
			} else {
				/*
				** Decode the action
				*/
				if ( !strcmp(action,"cd") || 
					 !strcmp(action,"get") || 
					 !strcmp(action,"getrfio") ||
					 !strcmp(action,"lcd") ||
					 !strcmp(action,"mget")  ||
					 !strcmp(action,"mkdir")  ||
					 !strcmp(action,"mput")  ||
					 !strcmp(action,"put") ||
					 !strcmp(action,"putrfio") ) {
					flagsigusr1 = 0 ;
					flagsigusr2 = 0 ;
					flagsighup = 0 ;
					fprintf(toserver,"%s\n",line)  ;
					fflush(toserver) ;
					while ( 1 == 1 ) {
						nfds = sysconf(_SC_OPEN_MAX) ;
						FD_ZERO(&selectmask) ;
						FD_SET(newstdout,&selectmask) ;
						if ( (retcode = select(nfds,&selectmask,0,0,0) ) == -1 ) {
							/*
							** If we got a signal we are here
							*/
							if (flagsigusr2 == 1 ) {
								fprintf(resfd,"%s FAILED\n",line) ;
								fflush(resfd) ;
								break ;
							} else if ( flagsighup == 1 ) {
								fprintf(resfd,"%s OK\n",line) ;
								fflush(resfd) ;
								break ;
							} 
						} else if ( retcode == 0 ) {

						} else {
							if ( FD_ISSET(newstdout,&selectmask) ) {
								if ( fgets(lineread,MAXLINE,fromserver) != NULL ) {
					 				printf("%s",lineread) ;
								}
							}
							if (flagsigusr2 == 1 ) {
								fprintf(resfd,"%s FAILED\n",line) ;
								fflush(resfd) ;
								break ;
							} else if ( flagsighup == 1 ) {
								fprintf(resfd,"%s OK\n",line) ;
								fflush(resfd) ;
								break ;
							}
						}
					}
				} else {
					printf("Unknown command : %s\n",line) ;
					fprintf(resfd,"%s FAILED\n",line) ;
					fflush(resfd) ;
				}
			}
		}
	}
#else
	printf("bbftpc is not available on this system\n") ;
	exit(1) ;
	
#endif
}
