/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 bbftpcd.c v 1.8.9  2000/08/01	- Program creation
 		   v 1.8.10 2000/08/11	- Portage to Linux
		   v 1.8.12 2000/08/14	- Add putrfio command
		   v 1.9.0  2000/08/14	- Add -r option

*****************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/un.h>
#include <unistd.h>
     

#include <bbftp.h>
#include <client.h>
#include <common.h>
#include <structures.h>
#include <version.h>
#define OPTIONS	"r:u:v"

struct	sockaddr_in data_source[MAXPORT] ;
int		datasock[MAXPORT] ; 
int		pid_child[MAXPORT] ;
int		debug = 0 ;
int		timestamp = 0 ;
FILE	*infd;
struct sockaddr_in hisctladdr ;
struct sockaddr_in myctladdr ;
char	*remotedir = NULL ;
char	*localdir = NULL ;
pid_t	fatherpid ;
int		unlinkfile ;
char	localfile[MAXLINE] ;
int		globaltrymax = NBTRYMAX ;


/*
** Specific to cd
*/
int		flagsigusr1 ;
pid_t	pidusr1 ;
uid_t	uidusr1 ;
int		clientpid ;

void destroyunixsoc() 
{
	struct	sockaddr_un unixsocaddr ;

	sprintf(unixsocaddr.sun_path,"/tmp/.bbftpcd.in.%d",clientpid) ;
	unlink(unixsocaddr.sun_path) ;
	sprintf(unixsocaddr.sun_path,"/tmp/.bbftpcd.out.%d",clientpid) ;
	unlink(unixsocaddr.sun_path) ;
}
void Usage() 
{
	printf("Usage bbftpcd -u username\n") ;
	printf("             [-r number of tries ]\n") ;
	printf("             host\n") ;
	printf("      or \n") ;
	printf("      bbftpcd -v\n") ;
	_exit(1) ;  
}

main(argc, argv, envp)
	int argc;
	char **argv;
	char **envp;
{
#ifndef HAVE_SIGINFO
	printf("bbftpcd is not available on this system\n") ;
	exit(1) ;
}
#else
	extern char *optarg;
	extern int optind, opterr, optopt;
	char	*username = NULL ;
	char	*password = NULL ;
	char	*hostname ;
	int		hosttype ;
	char	*calchostname ;
	struct hostent  *hp = NULL ;
	int		nbport = 1 ;
	int		compress = NOCOMPRESSION ;

	int		controlsock ;
	int		nbtry ;
	int		connectionisbroken ;
	struct	passwd	*uspass ;
	char tmpchar[50] ;
	struct	sockaddr_un unixsocaddr ;
	int		unixsocaddrlen ;
	struct timeval	wait_timer;

	int		nfds ;
	fd_set	selectmask ; /* Select mask */
	int		j, retcode, i ;
	
	int		newstdin,newstdout,newstderr ;
	int		tmpsock ;

	char	line[MAXLINE] ;
	int		nbitem ;
	char	action[MAXLINE] ;
	char	reste[MAXLINE] ;
	
	if ( blockallsignals() < 0 ) {
		printf("Error in blockallsignals function\n") ;
		_exit(1) ;
	}
	opterr = 1 ;
	while ((j = getopt(argc, argv, OPTIONS)) != -1) {
		switch (j) {
			case 'r' :{
				retcode = sscanf(optarg,"%d",&globaltrymax) ;
				if ( retcode != 1 ) {
					fprintf(stderr,"Number of tries must be numeric\n") ;
					_exit(1) ;
				}
				break ;
			}
			case 'u' :{
				username = optarg ;
				break ;
			}
			case 'v' :{
				printf("bbftpcd version %s compiled with %d tries\n",VERSION,NBTRYMAX) ;
				exit(0) ;
			}
			default : {
				Usage() ;
			}
		}
	}
	if ( optind != argc-1 ) Usage() ;
	hostname = argv[optind] ;
	hosttype = 0 ;

	for (i=0 ; i < strlen(hostname) ; i++) {
		if ( isalpha(hostname[i]) ) {
			hosttype = 1 ;
			break ;
		} else if ( isdigit(hostname[i]) ) {
		} else if ( hostname[i] == '.' ) {
		} else {
			printf("Invalid hostname : %s\n",hostname) ;
			_exit(1) ;
		}
	}
	if ( hosttype == 0 ) {
		/*
		** Adresse numerique
		*/
		hisctladdr.sin_addr.s_addr = 0 ;
		hisctladdr.sin_addr.s_addr = inet_addr(hostname) ;
		if (hisctladdr.sin_addr.s_addr == -1 ) {
			fprintf(stderr,"Incorrect address inet_addr %s\n",hostname) ;
			_exit(1) ;
		}
		calchostname = (char *)inet_ntoa(hisctladdr.sin_addr) ;
		if ( strcmp(hostname,calchostname) ) {
			fprintf(stderr,"Incorrect address %s\n",hostname) ;
			_exit(1) ;
		}
	} else {
		if ( (hp = gethostbyname(hostname) ) == NULL ) {
			/*
			*/
			printf("Host %s not found\n",hostname) ;
			_exit(1) ;
		} else {
			/*
			** C'est un nom de machine
 			*/
			if (hp->h_length > (int)sizeof(hisctladdr.sin_addr)) {
				hp->h_length = sizeof(hisctladdr.sin_addr);
			}
			memcpy(&hisctladdr.sin_addr, hp->h_addr_list[0], hp->h_length) ;
		}
	}
	if ( username == NULL ) {
		fprintf(stderr,"No username \n") ;
		Usage() ;
	}
/*
** Ask for password
*/
#ifdef USE_GETPASSPHRASE
	password = (char *) getpassphrase("Password: ") ; 
#else
	sprintf(line,"Password (max length = %d): ",sysconf(_SC_PASS_MAX)) ;
	password = (char *) getpass(line) ; 
#endif
	if ( strlen(password) == 0 ) {
		printf("You must enter a password\n") ;
		_exit(1) ;
	}
/*
** Connect to the remote host on control port
*/
	for ( nbtry = 1 ; nbtry <= globaltrymax ; nbtry++ ) {
		retcode = connecttoserver(username,password,&controlsock) ;
		if ( retcode < -1 ) {
			_exit(1) ;
		} else if ( retcode == 0 ) {
			break ;
		} else {
			if ( nbtry != globaltrymax ) {
				printf("Retrying connection waiting %d s\n",WAITRETRYTIME) ;
				sleep(WAITRETRYTIME) ;
			}
		}
	}
	if ( retcode != 0 && nbtry == globaltrymax+1) {
		printf("Maximum try reached (%d) aborting connection\n",globaltrymax) ;
		_exit(1) ;
	}
/*
** Connection was successfull so close it and start in backgroud
*/
	close(controlsock) ;
	retcode = fork() ;
	if ( retcode < 0 ) printf("Error forking\n") ;
	if (retcode > 0 ) exit(0) ;
	setsid() ;
/*
** Set the new signal actions
*/
	if ( setsignals_cd_grand_father() < 0 ) {
		printf("Error in setsignals_cd_grand_father function\n") ;
		_exit(1) ;
	}
/*
** For compatibility with all routines open infd on /dev/null
*/
	if ( (infd = fopen("/dev/null","r") ) == NULL) {
		printf("Unable to open /dev/null\n") ;
		_exit(1) ;
	}
/*
** Do not use stderr
*/
	if ( (newstderr	= open("/dev/null", O_RDWR)) != -1) {
		dup2(newstderr, STDERR_FILENO);
	}
	printf("Magic number is %d \n",getpid()) ;

loopin:
	nfds = sysconf(_SC_OPEN_MAX) ;
	flagsigusr1 = 0 ;
	if ( (retcode = select(nfds,0,0,0,0) ) == -1 ) {
		/*
		** Select error but that is the normal exit 
		** meaning we've got a signal
		*/		
		if ( flagsigusr1 == 1 && uidusr1 ==  getuid() ) {
			clientpid = pidusr1 ;
			flagsigusr1 = 0 ;
		} else {
			goto loopin ;
		}
	} else if ( retcode == 0 ) {
		/*
		** Time out
		*/
		goto loopin ;
	} else {
		/*
		** Normal exit if we were selectiong something
		*/
		goto loopin ;
	}
/*
** Receive a good signal from a good pid so disable USR1
** and TERM temporary
*/
	unsetsignals_cd_grand_father() ;
/*
** At this point we are going to fork
*/
	retcode = fork() ;
	if ( retcode < 0 ) {
		goto loopin ;
	}
	if (retcode > 0 ) {
		/*
		** In main magic number
		*/
		setsignals_cd_grand_father() ;
		goto loopin ;
	}
/*
** Now make the communications tools to get all commands
*/
	setsid() ;
	sprintf(tmpchar,"bbftpcd v%s",VERSION) ;
	openlog(tmpchar, LOG_PID | LOG_NDELAY, LOG_DAEMON);
	syslog(LOG_NOTICE,"Starting new connection for pid %d",clientpid) ;
	if ( (uspass =  getpwuid(getuid()) ) == NULL ) {
		syslog(LOG_ERR,"Error getting uspass : %d",getuid()) ;
		exit(1) ;
	}
	if ( uspass->pw_dir == NULL ) {
		syslog(LOG_ERR,"No home directory uid %d : %m",getuid()) ;
		exit(1) ;
	}
/*
** Create the unix socket 
*/
	if ( (newstdin = socket(AF_UNIX,SOCK_STREAM,0) ) < 0 ) {
		syslog(LOG_ERR,"Cannot create newstdin socket : %s",strerror(errno)) ;
		exit(1) ;
	}
	if ( (newstdout = socket(AF_UNIX,SOCK_STREAM,0) ) < 0 ) {
		syslog(LOG_ERR,"Cannot create newstdout socket : %s",strerror(errno)) ;
		exit(1) ;
	}
/*
** We use /tmp in order not to overflow unixsocaddr.sun_path
*/
/*
** For stdin
*/
	unixsocaddr.sun_family = AF_UNIX ;
	sprintf(unixsocaddr.sun_path,"/tmp/.bbftpcd.in.%d",clientpid) ;
	unixsocaddrlen = sizeof(unixsocaddr) ;
	if ( bind(newstdin,(struct sockaddr *)&unixsocaddr,unixsocaddrlen) < 0 ) {
		/*
		** That means that the file still exist which is completly
		** abnormal so destroy the file and retry
		*/
		unlink( unixsocaddr.sun_path) ;
		if ( bind(newstdin,(struct sockaddr *)&unixsocaddr,unixsocaddrlen) < 0 ) {
			/*
			** Again ... so exit 
			*/
			syslog(LOG_ERR,"Error binding stdin socket : %s",strerror(errno)) ;
			exit(1) ;
		}
	}
	if ( listen(newstdin,1) < 0 ) {
		destroyunixsoc() ;
		syslog(LOG_ERR,"Error listening stdin socket : %s",strerror(errno)) ;
		exit(1) ;
	}
/*
** For stdout
*/
	unixsocaddr.sun_family = AF_UNIX ;
	sprintf(unixsocaddr.sun_path,"/tmp/.bbftpcd.out.%d",clientpid) ;
	unixsocaddrlen = sizeof(unixsocaddr) ;
	if ( bind(newstdout,(struct sockaddr *)&unixsocaddr,unixsocaddrlen) < 0 ) {
		/*
		** That means that the file still exist which is completly
		** abnormal so destroy the file and retry
		*/
		unlink( unixsocaddr.sun_path) ;
		if ( bind(newstdout,(struct sockaddr *)&unixsocaddr,unixsocaddrlen) < 0 ) {
			/*
			** Again ... so exit 
			*/
			syslog(LOG_ERR,"Error binding stdout socket : %s",strerror(errno)) ;
			destroyunixsoc() ;
			exit(1) ;
		}
	}
	if ( listen(newstdout,1) < 0 ) {
		destroyunixsoc() ;
		syslog(LOG_ERR,"Error listening stdout socket : %s",strerror(errno)) ;
		exit(1) ;
	}
/*
** Send the USR1 signal to client to tell that unix sockets
** are ready 
*/
	/*
	** Initialize the pid array (for the case of SIGTERM)
	*/
	for ( i=0 ; i< MAXPORT ; i++) {
		pid_child[i] = 0 ;
	}
	fatherpid = getpid() ;
	unlinkfile = 0 ;
	if ( setsignals_cd_father() < 0 ) {
		printf("Error in setsignals_cd_father function\n") ;
		_exit(1) ;
	}
	kill(clientpid,STARTNEW) ;
/*
** Now wait for the connection with a time out
*/
/*
** stdin
*/
	nfds = sysconf(_SC_OPEN_MAX) ;
	FD_ZERO(&selectmask) ;
	FD_SET(newstdin,&selectmask) ;
	wait_timer.tv_sec  = 10 ;
	wait_timer.tv_usec = 0 ;
	if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
		destroyunixsoc() ;
		syslog(LOG_ERR,"Select error waiting connection stdin %s",strerror(errno)) ;
		exit(1) ;
	} else if ( retcode == 0 ) {
		/*
		** Time out, unlink all files and exit
		*/
		destroyunixsoc() ;
		syslog(LOG_ERR,"Time out waiting connection stdin") ;
		exit(1) ;
	} else {
		if ( (tmpsock = accept(newstdin,0,0) ) < 0 ) {
			destroyunixsoc() ;
			syslog(LOG_ERR,"Accept failed on stdin %s",strerror(errno)) ;
			exit(1) ;
		}
		close(newstdin) ;
		newstdin = tmpsock ;
	}
/*
** stdout
*/
	nfds = sysconf(_SC_OPEN_MAX) ;
	FD_ZERO(&selectmask) ;
	FD_SET(newstdout,&selectmask) ;
	wait_timer.tv_sec  = 10 ;
	wait_timer.tv_usec = 0 ;
	if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
		destroyunixsoc() ;
		syslog(LOG_ERR,"Select error waiting connection stdout %s",strerror(errno)) ;
		exit(1) ;
	} else if ( retcode == 0 ) {
		/*
		** Time out, unlink all files and exit
		*/
		destroyunixsoc() ;
		syslog(LOG_ERR,"Time out waiting connection stdout") ;
		exit(1) ;
	} else {
		if ( (tmpsock = accept(newstdout,0,0) ) < 0 ) {
			destroyunixsoc() ;
			syslog(LOG_ERR,"Accept failed on stdout %s",strerror(errno)) ;
			exit(1) ;
		}
		close(newstdout) ;
		newstdout = tmpsock ;
	}
/*
** Now make newstdin and newstdout stdin and out
*/
	if ( dup2(newstdin,STDIN_FILENO) < 0 ) {
		destroyunixsoc() ;
		syslog(LOG_ERR,"dup2 failed on stdin %s",strerror(errno)) ;
		exit(1) ;
		
	}
	if ( dup2(newstdout,STDOUT_FILENO) < 0 ) {
		destroyunixsoc() ;
		syslog(LOG_ERR,"dup2 failed on stdout %s",strerror(errno)) ;
		exit(1) ;
	}
/*
** Now we are ready to get the commands
** wait on the stdin 
*/
	if ( fgets(line,MAXLINE,stdin) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		/*
		** read the characteristics
		** We wait for 
		**       1) the current directory
		**       2) Compression (on /off)
		**       3) number of // streams
		**       4) debug mode (on/off)
		**       5) timestamp mode (on/off)
		**       6) number of try
		*/
		if ( chdir(line) < 0 ) {
			printf("Error changing dir to %s : %s\n",line,strerror(errno)) ;
			syslog(LOG_ERR,"Error changing dir to %s : %s",line,strerror(errno)) ;
			destroyunixsoc() ;
		 	exit(1) ;
		} else {
			syslog(LOG_NOTICE,"Changed dir to %s",line) ;
		}
	} else {
		syslog(LOG_NOTICE,"No change dir") ;
	}
	if ( fgets(line,MAXLINE,stdin) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		if ( !strcmp(line,"on") ) {
			compress = COMPRESSION ;
			syslog(LOG_NOTICE,"Compression on") ;
		} else {
			compress = NOCOMPRESSION ;
			syslog(LOG_NOTICE,"Compression off") ;
		}
	}
	if ( fgets(line,MAXLINE,stdin) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		retcode = sscanf(line,"%d",&nbport) ;
		if ( retcode != 1 ) {
			printf("Number of port must be numeric\n") ;
			destroyunixsoc() ;
			exit(1) ;
		}
		if ( nbport <= 0 || nbport > MAXPORT ) {
			printf("Number of port must be less than %d\n",MAXPORT) ;
			destroyunixsoc() ;
			exit(1) ;
		}
	}
	if ( fgets(line,MAXLINE,stdin) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		if ( !strcmp(line,"on") ) {
			debug  = 1 ;
			syslog(LOG_NOTICE,"Debug on") ;
		} else {
			debug = 0 ;
			syslog(LOG_NOTICE,"Debug off") ;
		}
	}
	if ( fgets(line,MAXLINE,stdin) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		if ( !strcmp(line,"on") ) {
			timestamp  = 1 ;
			syslog(LOG_NOTICE,"Timestamp on") ;
		} else {
			timestamp = 0 ;
			syslog(LOG_NOTICE,"Timestamp off") ;
		}
	}
	if ( fgets(line,MAXLINE,stdin) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		retcode = sscanf(line,"%d",&globaltrymax) ;
		if ( retcode != 1 ) {
			printf("Number of try must be numeric\n") ;
			destroyunixsoc() ;
			exit(1) ;
		}
	}
/*
** Connect to the remote host on control port
*/
	for ( nbtry = 1 ; nbtry <= globaltrymax ; nbtry++ ) {
		retcode = connecttoserver(username,password,&controlsock) ;
		if ( retcode < -1 ) {
			destroyunixsoc() ;
			exit(1) ;
		} else if ( retcode == 0 ) {
			break ;
		} else {
			if ( nbtry != globaltrymax ) {
				if (timestamp ) printtime() ;
				printf("Retrying connection waiting %d s\n",WAITRETRYTIME) ;
				sleep(WAITRETRYTIME) ;
			}
		}
	}
	if ( retcode != 0 && nbtry == globaltrymax+1) {
		if (timestamp ) printtime() ;
		printf("Maximum try reached (%d) aborting connection\n",globaltrymax) ;
		destroyunixsoc() ;
		exit(1) ;
	}
	connectionisbroken = 0 ;
/*
** Start the loop on stdin
*/
	while ( fgets(line,MAXLINE,stdin) != NULL ) {
		if ( line[strlen(line)-1] == '\n' ) line[strlen(line)-1] = '\0' ;
		nbitem = sscanf(line,"%s %s",action,reste) ;
		if (nbitem == EOF || nbitem == 0 ) {
			if (timestamp ) printtime() ;
			printf("Error on line : %s\n",line) ;
			kill(clientpid,TRANSFAIL) ;
		} else {
			/*
			** Decode the action
			*/
			if ( !strcmp(action,"cd") || 
				 !strcmp(action,"get") || 
				 !strcmp(action,"getrfio") ||
				 !strcmp(action,"lcd") ||
				 !strcmp(action,"mget")  ||
				 !strcmp(action,"mkdir")  ||
				 !strcmp(action,"mput")  ||
				 !strcmp(action,"put") ||
				 !strcmp(action,"putrfio") ) {
				for ( nbtry = 1 ; nbtry <= globaltrymax ; nbtry++ ) {
					/*
					** check if the last try was not disconnected
					*/
					/*
					** Initialize the pid array (for the case of SIGTERM)
					*/
					for ( i=0 ; i< MAXPORT ; i++) {
						pid_child[i] = 0 ;
					}
					if ( connectionisbroken == 1 ) {
						reconnecttoserver(username,password,&controlsock) ;
						connectionisbroken = 0 ;
					}
					if (!strcmp(action,"cd") ) {
						retcode = changedir(line,controlsock,nbport,compress) ;
					} else if (!strcmp(action,"get") ) {
						unlinkfile = 0 ;
						retcode = getafile(line,controlsock,nbport,compress,0) ;
						unlinkfile = 0 ;
					} else if (!strcmp(action,"getrfio") ) {
						unlinkfile = 0 ;
						retcode = getafile(line,controlsock,nbport,compress,1) ;
						unlinkfile = 0 ;
					} else if (!strcmp(action,"lcd") ) {
						retcode = localchangedir(line) ;
					} else if (!strcmp(action,"mget") ) {
						retcode = mgetfile(line,&controlsock,nbport,compress,stderr,username,password) ;
					} else if (!strcmp(action,"mkdir") ) {
						retcode = makeadir(line,controlsock) ;
					} else if (!strcmp(action,"mput") ) {
						retcode = mputfile(line,&controlsock,nbport,compress,stderr,username,password) ;
					} else if (!strcmp(action,"put") ) {
						retcode = putafile(line,controlsock,nbport,compress) ;
					} else if (!strcmp(action,"putrfio") ) {
						retcode = putafilerfio(line,controlsock,nbport,compress) ;
					}
					if ( retcode == BB_RET_FT_NR ) {
						/*
						** Fatal error no retry and connection is not broken
						*/
						break ;
					} else if ( retcode == BB_RET_FT_NR_CONN_BROKEN ) {
						/*
						** Fatal error no retry and connection is broken
						*/
						connectionisbroken = 1 ;
						break ;
					} else if ( retcode == BB_RET_OK ) {
						/*
						** Success , no retry
						*/							
						break ;
					} else if ( retcode == BB_RET_CONN_BROKEN ) {
						/*
						** Retry needed with a new connection
						*/
						if ( nbtry != globaltrymax ) {
							if (timestamp ) printtime() ;
							printf("Retrying connection waiting %d s\n",WAITRETRYTIME) ;
							sleep(WAITRETRYTIME) ;
							reconnecttoserver(username,password,&controlsock) ;
						} else {
							/*
							** In case of broken connection and max retry reached
							** indicate to the next transfer that the connection
							** is broken
							*/
							connectionisbroken = 1 ;
						}
					} else {
						/*
						** retcode > 0 means retry on this transfer
						*/
						if ( nbtry != globaltrymax ) {
							if (timestamp ) printtime() ;
							printf("Retrying transfer waiting %d s\n",WAITRETRYTIME) ;
							sleep(WAITRETRYTIME) ;
						}
					}
				}
				if ( retcode != 0 && nbtry == globaltrymax+1) {
					if (timestamp ) printtime() ;
					printf("Maximum try reached (%d) skipping file\n",globaltrymax) ;
					kill(clientpid,TRANSFAIL) ;
				} else if ( retcode == 0 ) {
					fflush(stdout) ;
					kill(clientpid,TRANSOK) ;
				} else {
					kill(clientpid,TRANSFAIL) ;
				}
			} else {
				if (timestamp ) printtime() ;
				printf("Unknown command : %s\n",line) ;
				kill(clientpid,TRANSFAIL) ;
			}
		}
	}
	destroyunixsoc() ;
}
#endif 

#ifndef WORDS_BIGENDIAN
my64_t convertlong(my64_t v) {
	struct bb {
		int	fb ;
		int sb ;
	} ;
	struct bb *bbpt ;
	int 	tmp ;
	my64_t	tmp64 ;
	
	tmp64 = v ;
	bbpt = (struct bb *) &tmp64 ;
	tmp = bbpt->fb ;
	bbpt->fb = ntohl(bbpt->sb) ;
	bbpt->sb = ntohl(tmp) ;	
	return tmp64 ;
}
#endif 
