/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN :
 	 BB_RET_OK                  = everything OK 
	 BB_RET_CONN_BROKEN         = connection with server broken retry with connection first
	 BB_RET_FT_NR_CONN_BROKEN   = Fatal error no retry needed on this command and connection is
	                              broken
	 BB_RET_FT_NR               = Fatal error no retry needed on this command but connection has been
	                              kept
	 BB_RET_ERROR               = Error but retry possible
	   
 changedir.c v 1.8.8  2000/06/02	- Creation of the routine 
		  	 v 1.8.9  2000/08/08	- Add timestamp
		  	 v 1.8.10 2000/08/10	- Portage to Linux
 			 v 1.9.0 2000/08/21	- Use configure to help portage
 
*****************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <bbftp.h>
#include <client.h>
#include <common.h>
#include <structures.h>

extern char *remotedir ;
extern int		timestamp;

int changedir(char *line, int controlsock) 
{
	int	nbitem ; /* Number of item in line, must be equal to 2 */
	char	action[MAXLINE] ;
	char	remotefile[MAXLINE] ;
	int		code ;
	int		retcode ;
	int		msglen ;
	
	
	fd_set	selectmask ; /* Select mask */
	int	nfds ; /* Max number of file descriptor */
	
	char	send_buffer[MAXMESSLEN] ;
	char	answer_buffer[MAXMESSLEN] ;
 	struct	message *msg ;
	char	*tmpdir ;

	
	nbitem = sscanf(line,"%s %s",action,remotefile) ;
	
	if (nbitem != 2 ) {
		if (timestamp ) printtime() ;
		printf("Incorrect cd command : %s\n",line) ;
		return BB_RET_FT_NR ;
	}
	if (timestamp ) printtime() ;
	printf(">> COMMAND : %s\n",line) ;
	/*
	** Now send the control message 
	*/
	/* 
	** First part
	*/
	msg = (struct message *)send_buffer ;
	msg->code = MSG_CHDIR ;
#ifndef WORDS_BIGENDIAN
	msg->msglen = ntohl(strlen(remotefile)) ;
#else
	msg->msglen = strlen(remotefile) ;
#endif
	if ( writemessage(controlsock,send_buffer,MINMESSLEN,SENDCONTROLTO,0) < 0 ) {
		/*
		** We were not able to send the minimum message so
		** we are going to close the control socket and to 
		** tell the calling program to restart a connection
		*/
		if (timestamp ) printtime() ;
		printf("Error sending cd message part 1 : %s\n",strerror(errno)) ;
		close(controlsock) ;
		return BB_RET_CONN_BROKEN ; /* restart connection */
	}
	/* 
	** Second part
	*/
	if ( writemessage(controlsock,remotefile,strlen(remotefile),SENDCONTROLTO,0) < 0 ) {
		if (timestamp ) printtime() ;
		printf("Error sending cd message part 2 : %s\n",strerror(errno)) ;
		close(controlsock) ;
		return BB_RET_CONN_BROKEN ; /* restart connection */
	}
	/*
	** Now we are going to wait for the message on the control 
	** connection
	*/
waitcontrol:
	nfds = sysconf(_SC_OPEN_MAX) ;
	FD_ZERO(&selectmask) ;
	FD_SET(controlsock,&selectmask) ;
	retcode = select(nfds,&selectmask,0,0,0) ;
	if ( retcode < 0 ) {
		/*
		** Select error
		*/
		if ( errno != EINTR ) {
			/*
			** we have got an error so close the connection
			** and restart
			*/
			if (timestamp ) printtime() ;
			printf("Error select waiting on control connection: %s\n ",strerror(errno) ) ;
			close(controlsock) ;
			return BB_RET_CONN_BROKEN ;
		} else {
			/*
			** Interrupted by a signal
			*/
			FD_ZERO(&selectmask) ;
			FD_SET(controlsock,&selectmask) ;
			goto waitcontrol ;
		}
	} else if ( retcode == 0 ) {
		/*
		** Impossible we do not set any timer
		*/
		FD_ZERO(&selectmask) ;
		FD_SET(controlsock,&selectmask) ;
		goto waitcontrol ;
	} else {
		/*
		** read the message
		*/
		if ( readmessage(controlsock,answer_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			if (timestamp ) printtime() ;
			printf("Error reading control message after waitcontrol\n") ;
			close(controlsock) ; 
			return BB_RET_CONN_BROKEN ;
		}
		msg = (struct message *)answer_buffer ;
		code = msg->code ;
		if ( code == MSG_BAD || code == MSG_BAD_NO_RETRY) {
			/*
			** The directory creation has failed so ...
			*/
#ifndef WORDS_BIGENDIAN
			msglen = ntohl(msg->msglen) ;
#else
			msglen = msg->msglen ;
#endif
			if ( readmessage(controlsock,answer_buffer,msglen,CONTROLSOCKTO,0) < 0 ) {
				if (timestamp ) printtime() ;
				printf("Error reading BAD message after send waitcontrol\n") ;
				close(controlsock) ;
				if ( code == MSG_BAD ) {
					return BB_RET_CONN_BROKEN ;
				} else {
					return BB_RET_FT_NR_CONN_BROKEN ;
				}
			} else {
				answer_buffer[msglen] = '\0' ;
				if (timestamp ) printtime() ;
				printf("<< %s\n",answer_buffer) ;
				if ( code == MSG_BAD ) {
					return BB_RET_ERROR;
				} else {
					/*
					** In case of no retry we are going to wait 2 seconds
					** in order for the server to finish its cleaning (in
					** fact if the error message is du to a server child
					** abnormal termination, the message is sent on the death
					** of the first child and the server has some cleaning
					** to do).
					*/
					sleep(2) ;
					return BB_RET_FT_NR ;
				}
			}
		} else if (msg->code == MSG_OK ) {
			/*
			** At this stage the transfer is OK but we need
			** the answer because it gives the directory
			*/
#ifndef WORDS_BIGENDIAN
			msglen = ntohl(msg->msglen) ;
#else
			msglen = msg->msglen ;
#endif
			if ( readmessage(controlsock,answer_buffer,msglen,CONTROLSOCKTO,0) < 0) {
				if (timestamp ) printtime() ;
				printf("Error reading BAD message after send waitcontrol\n") ;
				close(controlsock) ;
				return BB_RET_CONN_BROKEN ;
			} else {
				if ( (tmpdir = (char *) malloc(msglen+1) ) == NULL ) {
					if (timestamp ) printtime() ;
					printf("Unable to malloc memory to keep track of directory\n") ;
					close(controlsock) ;
					return BB_RET_CONN_BROKEN ;
				} else {
					strncpy(tmpdir,answer_buffer,msglen) ;
					free(remotedir) ;
					remotedir = tmpdir ;
					remotedir[msglen] = '\0' ;
				}
			}
			if (timestamp ) printtime() ;
			printf("<< OK : Current remote directory is %s\n",remotedir) ;
			return BB_RET_OK ;
		} else {
			/*
			** Receive unkwown message so something is
			** going wrong. close the control socket
			** and restart
			*/
			if (timestamp ) printtime() ;
			printf("Receive unkown message waitcontrol : %d\n",msg->code) ;
			close(controlsock) ; 
			return BB_RET_CONN_BROKEN ;
		}
	}
}
