/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN :
 	 0    = evrything OK connection and login OK
	 -1   = connection with server broken retry 
	 <-1  = Fatal error (Login incorrect or unrecoverable error)
	   
 connecttoserver.c v 1.5.0  2000/03/23 	- Creation of the routine
 				   v 1.6.1  2000/03/28	- Print code in hexa for debuging
				   						  with OSF
				   v 1.7.1  2000/04/04	- Define hisctladdr and myctladdr
				   						  as extern (detected on SGI)
				   v 1.8.0  2000/04/14	- Introduce RSA Cryptage
				   v 1.8.3  2000/04/18	- Add the MESS_BAD_NO_RETRY message
				   v 1.8.7  2000/05/28	- Portage to OSF1
				   						- Set socket in non blocking mode
										- Change headers
		  		   v 1.8.9  2000/08/08	- Add timestamp
		  		   v 1.8.10 2000/08/10	- Portage to Linux
				   v 1.9.0  2000/08/21	- Use configure to help portage


*****************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <bbftp.h>
#include <config.h>
#include <common.h>
#include <structures.h>
#include <openssl/rsa.h>

extern struct sockaddr_in hisctladdr ;
extern struct sockaddr_in myctladdr ;
extern int	timestamp ;

int connecttoserver(char *username, char *password, int *controlsock) 
{
	int 	addrlen ;
	int		tmpctrlsock ;
	
	char	answer_buffer[MAXMESSLEN] ;
	char	buffer[MAXMESSLEN] ;
	char	buffer_sup[MAXMESSLEN] ;
	struct	message *msg ;
	struct	mess_sec *msg_sec ;
	struct	mess_rsa *msg_rsa ;
	int		msglen ;
	int		crtype ;
	int		code ; 

	char	buffrand[NBITSINKEY] ;
	unsigned char	pubkey[NBITSINKEY] ;
	unsigned char	pubexponent[NBITSINKEY] ;
	int		lenkey ;
	int		lenexpo ;
	unsigned int seed ;
	struct timeval tp ;
	int	i ;
	
	RSA		*hisrsa ;
	int lenrsa ;
	/*
	** Get the socket
	*/
	
	hisctladdr.sin_family = AF_INET;
	hisctladdr.sin_port = htons(CONTROLPORT);
	if ( (tmpctrlsock = socket ( AF_INET, SOCK_STREAM, IPPROTO_TCP )) < 0 ) {
		if (timestamp ) printtime() ;
		printf( "Cannot create control socket : %s",strerror(errno));
		return -1 ;
	}
	/*
	** Connect to the server
	*/
	addrlen = sizeof(hisctladdr) ;
	if ( connect(tmpctrlsock,(struct sockaddr*)&hisctladdr,addrlen) < 0 ) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf( "Cannot connect to control socket: %s\n",strerror(errno));
		return -1 ;
	}
	/*
	** Get the socket name
	*/
	addrlen = sizeof(myctladdr) ;
	if (getsockname(tmpctrlsock,(struct sockaddr*) &myctladdr, &addrlen) < 0) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Error getsockname on control socket: %s\n",strerror(errno)) ;
		return -1 ;
	}
	/*
	** Set the socket as non blocking
	*/
	if ( fcntl(tmpctrlsock,F_SETFL,O_NONBLOCK)  < 0 ) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Cannot fcntl O_NONBLOCK on control socket : %s\n",strerror(errno)) ;
		exit(1) ;
	}
	/*
	** Connection is correct so start the login procedure
	*/
	/*
	**	Read the encryption supported
	*/
	if ( readmessage(tmpctrlsock,answer_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Error reading encryption message\n") ;
		return -1 ;
	}
	msg = (struct message *) answer_buffer ;
	if ( msg->code != MSG_CRYPT) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("No encryption message : %08x\n",msg->code) ;
		return -1 ;
	}
	if ( readmessage(tmpctrlsock,answer_buffer,CRYPTMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Error reading encryption type message\n") ;
		return -1 ;
	}

	msg_sec = (struct mess_sec	*) answer_buffer ;
	if ( (msg_sec->crtype & CRYPT_RSA_PKCS1_OAEP_PADDING ) == CRYPT_RSA_PKCS1_OAEP_PADDING) {
		/* 
		** RSA
		*/
		/*
		** Load the error message from the crypto lib
		*/
		ERR_load_crypto_strings() ;
		/*
			** Initialize the buffrand buffer which is giong to be used to initialize the 
		** random generator
		*/
		/*
		** Take the usec to initialize the random session
		*/
		gettimeofday(&tp,NULL) ;
		seed = tp.tv_usec ;
		srandom(seed) ;
		for (i=0; i < sizeof(buffrand) ; i++) {
			buffrand[i] = random() ;
		}
		/*
		** Initialize the random generator
		*/
		RAND_seed(buffrand,NBITSINKEY) ;
#ifndef WORDS_BIGENDIAN
		lenkey = ntohl(msg_sec->pubkeylen) ;
		lenexpo = ntohl(msg_sec->expolen) ;
#else
		lenkey = msg_sec->pubkeylen ;
		lenexpo = msg_sec->expolen ;
#endif
		/*
		** Read the public key
		*/
		if ( readmessage(tmpctrlsock,pubkey,lenkey,CONTROLSOCKTO,0) < 0 ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error reading public key\n") ;
			return -1 ;
		}
		/*
		** Read the public exponent
		*/
		if ( readmessage(tmpctrlsock,pubexponent,lenexpo,CONTROLSOCKTO,0) < 0 ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error reading public exponent\n") ;
			return -1 ;
		}
		if ( (hisrsa = RSA_new()) == NULL) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error getting RSA: %s\n",(char *) ERR_error_string(ERR_get_error(),NULL) ) ;
			return -1 ;
		}
		/*
		** Getting BIGNUM structures to store the key and exponent
		*/
		if ( (hisrsa->n = BN_new()) == NULL) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error getting BIGNUM: %s\n",(char *)ERR_error_string(ERR_get_error(),NULL) ) ;
			return -1 ;
		}
		if ( (hisrsa->e = BN_new()) == NULL) { 
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error getting BIGNUM: %s\n",(char *)ERR_error_string(ERR_get_error(),NULL) ) ;
			return -1 ;
		}
		/*
		** Copy the key and exponent received
		*/
		if ( BN_mpi2bn(pubkey,lenkey,hisrsa->n) == NULL ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error copying pubkey: %s\n",(char *) ERR_error_string(ERR_get_error(),NULL) ) ;
			return -1 ;
		}
		if ( BN_mpi2bn(pubexponent,lenexpo,hisrsa->e) == NULL ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error copying pubexponent: %s\n",(char *) ERR_error_string(ERR_get_error(),NULL) ) ;
			return -1 ;
		}
		lenrsa = RSA_size(hisrsa) ;
		if (strlen(username) > lenrsa - 21 ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Unable to crypt username (len too long) (%d/%d)\n",strlen(username),lenrsa-21 ) ;
			return -2 ;
		}
		if (strlen(password) > lenrsa - 21 ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Unable to crypt password (len too long) (%d/%d)\n",strlen(password),lenrsa-21 ) ;
			return -2 ;
		}
		msg_rsa = ( struct mess_rsa *) buffer ;
		if ( (msg_rsa->numuser = RSA_public_encrypt(strlen(username),(unsigned char *)username,msg_rsa->cryptuser,hisrsa,RSA_PKCS1_OAEP_PADDING)) < 0 ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("%s\n",(char *)ERR_error_string(ERR_get_error(),NULL) ) ;
			return -1 ;
		}
#ifndef WORDS_BIGENDIAN
		msg_rsa->numuser = ntohl(msg_rsa->numuser) ;
#endif
		if ( (msg_rsa->numpass = RSA_public_encrypt(strlen(password),(unsigned char *)password,msg_rsa->cryptpass,hisrsa,RSA_PKCS1_OAEP_PADDING)) < 0 ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("%s\n",(char *)ERR_error_string(ERR_get_error(),NULL) ) ;
			return -1 ;
		}
#ifndef WORDS_BIGENDIAN
		msg_rsa->numpass = ntohl(msg_rsa->numpass) ;
#endif
		crtype = CRYPT_RSA_PKCS1_OAEP_PADDING ;
	} else {
		if (timestamp ) printtime() ;
		printf("Unkown encryption method %08x\n",msg_sec->crtype) ;
		return -1  ;
	}
	/*
	** Send login information
	*/
	/*
	** First message code 
	*/
	msg = (struct message *)buffer_sup ;
	msg->code = MSG_LOG ;
#ifndef WORDS_BIGENDIAN
	msg->msglen = ntohl(CRYPTMESSLEN+RSAMESSLEN) ;
#else
	msg->msglen = CRYPTMESSLEN+RSAMESSLEN ;
#endif
	if ( writemessage(tmpctrlsock,buffer_sup,MINMESSLEN,0) < 0) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Error sending username part 1 : %s\n",strerror(errno)) ;
		return -1 ;
	}
	/*
	** Then crypte type
	*/
	msg_sec = (struct mess_sec *)buffer_sup ;
	msg_sec->crtype = crtype ;
	if ( writemessage(tmpctrlsock,buffer_sup,CRYPTMESSLEN,0) < 0 ) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Error sending username part 2 : %s\n",strerror(errno)) ;
		return -1 ;
	}
	/*
	** Then real informations
	*/
	if ( writemessage(tmpctrlsock,buffer,RSAMESSLEN,0) < 0) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Error sending username part 3 : %s\n",strerror(errno)) ;
		return -1 ;
	}
	if (timestamp ) printtime() ;
	printf(">> USERNAME %s PASS\n",username) ;
	/*
	** Now wait for the OK message
	*/
	if ( readmessage(tmpctrlsock,answer_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Error reading login message answer\n") ;
		return -1 ;
	}
	msg = (struct message *)answer_buffer ;
	code = msg->code ;
	if ( code == MSG_BAD || code == MSG_BAD_NO_RETRY) {
#ifndef WORDS_BIGENDIAN
		msglen = ntohl(msg->msglen) ;
#else
		msglen = msg->msglen ;
#endif
		if ( readmessage(tmpctrlsock,answer_buffer,msglen,CONTROLSOCKTO,0) < 0 ) {
			close(tmpctrlsock) ;
			if ( code == MSG_BAD ) {
				if (timestamp ) printtime() ;
				printf("Error reading BAD message on login\n") ;
				return -1 ;
			} else {
				if (timestamp ) printtime() ;
				printf("Error reading BAD_NO_RETRY message on login\n") ;
				return -2 ;
			}
		} else {
			close(tmpctrlsock) ;
			answer_buffer[msglen] = '\0' ;
			if (timestamp ) printtime() ;
			printf("<< %s\n",answer_buffer) ;
			if ( code == MSG_BAD ) {
				return -1 ;
			} else {
				return -2 ;
			}
		}
	} else if ( code == MSG_OK ) {
#ifndef WORDS_BIGENDIAN
		msglen = ntohl(msg->msglen) ;
#else
		msglen = msg->msglen ;
#endif
		if ( readmessage(tmpctrlsock,answer_buffer,msglen,CONTROLSOCKTO,0) < 0 ) {
			close(tmpctrlsock) ;
			if (timestamp ) printtime() ;
			printf("Error reading OK message on login\n") ;
			return -1 ;
		} else {
			answer_buffer[msglen] = '\0' ;
			if (timestamp ) printtime() ;
			printf("<< %s\n",answer_buffer) ;
		}		
	} else {
		close(tmpctrlsock) ;
		if (timestamp ) printtime() ;
		printf("Unkown answer message at login : %08x\n",code) ;
		return -1 ;
	}
	
	*controlsock = tmpctrlsock ;
	return 0 ;
	
}
