/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN :
 	 0    = everything OK file correctly transfered
	 -1   = connection with server broken retry with connection first
	 -2   = Fatal error no retry needed on this command and connection is
	        broken
	 <-2  = Fatal error no retry needed on this command but connection has been
	        kept
	 >0   = Error but retry possible
	   
 getafile.c v 1.6.0  2000/03/24	- Creation of the routine
 			v 1.7.1  2000/04/04	- Correct printf on error (detected on SGI)
			v 1.8.0  2000/04/14	- Change location of headers file for zlib
			v 1.8.3  2000/04/18	- Implement the MSG_BAD_NO_RETRY message and
								  change the return code (-2 added)
			v 1.8.7  2000/05/28	- Portage to OSF1
				   				- Set socket in non blocking mode
								- Change headers
			v 1.8.8  2000/05/30	- Check if file to create is not a directory
								- Accept get command with one filename
			  					- Take care of a zero length file
		    v 1.8.9  2000/08/08	- Add timestamp
								- Add Child check in waitcontrol
		    v 1.8.10 2000/08/10	- Portage to Linux
			v 1.9.0  2000/08/21	- Use configure to help portage

*****************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <limits.h>

#include <bbftp.h>
#include <client.h>
#include <common.h>
#include <structures.h>

#include <zlib/zlib.h>

#ifndef WORDS_BIGENDIAN
my64_t	convertlong(my64_t v) ;
#endif

extern struct sockaddr_in data_source[MAXPORT] ;
extern struct sockaddr_in hisctladdr ;
extern struct sockaddr_in myctladdr ;
extern int		datasock[MAXPORT] ; 
extern int		pid_child[MAXPORT] ;
extern int		debug ;
extern int		timestamp ;
extern	FILE	*infd ;
extern char		localfile[MAXLINE] ;
extern int		unlinkfile ;

int getafile(char *line, int controlsock, int nbport, int compresstype, int mode) 
{
	int		nbitem ; /* Number of item in line, must be equal to 3 or 2*/
	char	action[MAXLINE] ;
	char	remotefile[MAXLINE] ;
	
	int		i,retcode,k ;
	int		code ;
	int		status ;
	pid_t	pid ;

	int		msglen ;
	char	send_buffer[MAXMESSLEN] ;
	char	receive_buffer[MAXMESSLEN] ;
 	struct	message *msg ;
	struct	mess_store *msg_store ;
	struct	mess_retr_ok *msg_retr_ok ;
	struct	mess_compress *msg_compress ;
	
	int		realnbport ;
	int		ns ;
	
	fd_set	selectmask ; /* Select mask */
	int		nfds ; /* Max number of file descriptor */
	struct timeval	wait_timer;

	int		fd ;

	my64_t	filelen ;

#ifdef STANDART_FILE_CALL
	off_t	startpoint ;
	off_t	nbperchild ;
	off_t 	nbtoget ;
	off_t 	nbget ;
	off_t	toseek ;
	struct stat statbuf ;
#else
	off64_t		startpoint ;
	off64_t		nbperchild ;
	off64_t 	nbtoget ;
	off64_t 	nbget ;
	off64_t		toseek ;
	struct stat64 statbuf	;
#endif

	int	 	datatoreceive ;
	int	 	dataonone ;

	char	data_buffer[READBUFLEN] ;
	char	comp_buffer[READBUFLEN] ;
	int	compressionon ;
	uLong	buflen ;
	uLong	bufcomplen ;
	size_t lentowrite;
	size_t lenwrited;

	struct timeval  tstart, tend ,tdiff;
	float	s, bs;
	int		savederrno ;
		
	/*
	** Initialize the pid array (for the case of SIGTERM)
	*/
	for ( i=0 ; i< MAXPORT ; i++) {
		pid_child[i] = 0 ;
	}
	realnbport = nbport ;
	
	nbitem = sscanf(line,"%s %s %s",action,remotefile,localfile) ;

	if (nbitem == 2 ) {
		/*
		** This mean that localfile = remotefile
		*/
		strcpy(localfile,remotefile) ;
	} else if (nbitem != 3 ) {
		if (timestamp ) printtime() ;
		printf("Incorrect get command : %s \n",line) ;
		return -3 ;
	}
	if (timestamp ) printtime() ;
	printf(">> COMMAND : %s\n",line) ;
	/*
	** Before doing anything we are going to check if the local
	** file is not a directory
	*/
#ifdef STANDART_FILE_CALL
	if ( stat(localfile,&statbuf) < 0 ) {
#else
	if ( stat64(localfile,&statbuf) < 0 ) {
#endif
		/*
		** It may be normal to get an error if the file
		** does not exist but some error code must lead
		** to the interruption of the transfer:
		**		EACCES		: Search permission denied
		**		ELOOP		: To many symbolic links on path
		**		ENAMETOOLONG: Path argument too long
		**		ENOTDIR		: A component in path is not a directory
		*/
		savederrno = errno ;
		if ( savederrno == EACCES ||
			savederrno == ELOOP ||
			savederrno == ENAMETOOLONG ||
			savederrno == ENOTDIR ) {
			if (timestamp ) printtime() ;
			printf("Error stating file %s : %s \n",localfile,strerror(savederrno)) ;
			return -3 ;
		} else if (savederrno == ENOENT) {
			/*
			** That is normal the file does not exist
			*/
		} else {
			if (timestamp ) printtime() ;
			printf("Error stating file %s : %s \n",localfile,strerror(savederrno)) ;
			return 1 ;
		}
	} else {
		/*
		** The file exists so check if it is a directory
		*/
		if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
			if (timestamp ) printtime() ;
			printf("File %s is a directory\n",localfile) ;
			return -3 ;
		}
	}				
	/*
	** Clear the data_source structure
	*/
	for (i=0 ; i< MAXPORT ; i++) {
		data_source[i] = myctladdr ;
		data_source[i].sin_port = 0;
		datasock[i] = 0 ;
		pid_child[i] = 0 ;
	}
	/*
	** Get the data socket
	*/
	if ( getdatasock(realnbport) < 0 ) {
		if (timestamp ) printtime() ;
		printf("Error getting data socket\n") ;
		/*
		** In case of error getdatasock does not clean the
		** structure and does not close the sockets, it is
		** the calling program's work
		*/ 
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		/*
		** In this case we allow retry
		*/
		return 1 ;
	}
	/*
	** We cannot start the children because we are not able to know the file 
	** size so we send the request
	*/
	/* 
	** First part
	*/
	msg = (struct message *)send_buffer ;
	if ( compresstype == NOCOMPRESSION ) {
		if ( mode == 0 ) msg->code = MSG_RETR ;
		if ( mode == 1 ) msg->code = MSG_RETR_RFIO ;
	} else {
		if ( mode == 0 ) msg->code = MSG_RETR_C ;
		if ( mode == 1 ) msg->code = MSG_RETR_RFIO_C ;
	}
#ifndef WORDS_BIGENDIAN
	msg->msglen = ntohl(STORMESSLEN) ;
#else
	msg->msglen = STORMESSLEN ;
#endif
	if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		/*
		** Fail to send
		** Close the connection and ask for a restart of connexion
		*/
		if (timestamp ) printtime() ;
		printf("Error sending retr message part 1 \n") ;
		close (controlsock) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		return -1 ; /* restart connection */
	}
	msg_store = (struct mess_store *)send_buffer ;
	strcpy(msg_store->filename,remotefile)  ;
#ifndef WORDS_BIGENDIAN
	msg_store->filesize = 0 ;
	msg_store->nbport = ntohl(realnbport) ;
#else
	msg_store->filesize = 0 ;
	msg_store->nbport = realnbport ;
#endif
	for (i=0 ; i< realnbport ; i++) {
#ifndef WORDS_BIGENDIAN
		msg_store->port[i]  = ntohl(ntohs(data_source[i].sin_port)) ;
#else
		msg_store->port[i]  = data_source[i].sin_port ;
#endif
		if ( debug ) printf("Using Port number = %d\n",data_source[i].sin_port) ;
	}
	if ( writemessage(controlsock,send_buffer,STORMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		/*
		** Fail to send
		** Close the connection and ask for a restart of connexion
		*/
		if (timestamp ) printtime() ;
		printf("Error sending retr message part 2 \n") ;
		close (controlsock) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		return -1 ; /* restart connection */
	}
	/*
	** Now we are going to wait on the control sock in order
	** to retrieve the file length we set up a timer to be able
	** to leave
	*/
	wait_timer.tv_sec  = GETANSWAITTIME ;
	wait_timer.tv_usec = 0 ;
	nfds = sysconf(_SC_OPEN_MAX) ;
	FD_ZERO(&selectmask) ;
	FD_SET(controlsock,&selectmask) ;
	retcode = select(nfds,&selectmask,0,0,&wait_timer) ;
	if ( retcode < 0 ) {
		/*
		** select return in error 
		** Leave and break the connection
		*/
		if (timestamp ) printtime() ;
		printf("Error select waiting answer to get message %s\n",strerror(errno)) ;
		close (controlsock) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		return -1 ;/* restart connection */
	}
	if ( retcode == 0 ) {
		/*
		** That is a time out 
		*/
		if (timestamp ) printtime() ;
		printf("Time out on select waiting answer to get messag\n") ;
		close (controlsock) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		return -1 ;/* restart connection */
	}
	/*
	** Read the message
	*/
	if ( readmessage(controlsock,receive_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		/*
		** Error reading the message
		*/
		if (timestamp ) printtime() ;
		printf("Error reading answer to get message \n") ;
		close (controlsock) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		return -1 ;/* restart connection */
	}
	msg = (struct message *)receive_buffer ;
	code = msg->code ;
	if ( code == MSG_BAD || code == MSG_BAD_NO_RETRY ) {
			/*
			** Something went wrong ... The return code has to be
			** improved in future releases
			*/
#ifndef WORDS_BIGENDIAN
		msglen = ntohl(msg->msglen) ;
#else
		msglen = msg->msglen ;
#endif
		if ( readmessage(controlsock,receive_buffer,msglen,CONTROLSOCKTO,0) < 0 ) {
			if (timestamp ) printtime() ;
			printf("Error reading BAD message after answer\n") ;
			close (controlsock) ;
			/*
			** Close all other sockets
			*/
			for (i=0 ; i<realnbport ; i++) {
				if ( datasock[i] != 0 ) {
					close(datasock[i]) ;
					datasock[i] = 0 ;
				}
			}
			if ( code == MSG_BAD ) {
				return -1 ;
			} else {
				return -2 ;
			}
		}
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		receive_buffer[msglen] = '\0' ;
		if (timestamp ) printtime() ;
		printf("<< %s\n",receive_buffer) ;
		if ( code == MSG_BAD ) {
			return 1;
		} else {
			return -3 ;
		}
	} else if (code != MSG_RETR_OK ) {
		if (timestamp ) printtime() ;
		printf("Unkown answer message %d\n",msg->code) ;
		close (controlsock) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		return -1 ;/* restart connection */
	}
	/*
	** Read the file length
	*/
	if ( readmessage(controlsock,receive_buffer,RETROKMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		/*
		** Error reading the message
		*/
		if (timestamp ) printtime() ;
		printf("Error reading file length\n") ;
		close (controlsock) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		return -1 ;/* restart connection */
	}
	msg_retr_ok = (struct mess_retr_ok *)receive_buffer ;
#ifndef WORDS_BIGENDIAN
	filelen = convertlong(msg_retr_ok->filesize) ;
#else
	filelen = msg_retr_ok->filesize ;
#endif
	if (debug) printf("received file length : %" LONG_LONG_FORMAT "\n",filelen) ;
#ifndef ALLOW_64BITS_FILES
	if ( filelen > INT_MAX ) {
		if (timestamp ) printtime() ;
		printf("File size too big : %" LONG_LONG_FORMAT " (max is %d)\n",filelen,INT_MAX) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		
		msg = (struct message *)send_buffer ;
		msg->code = MSG_ABR ;
		msg->msglen = 0 ;
		if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			/*
			** Error sending the Abort message
			** Close the connection and ask for a restart of connexion
			*/
			if (timestamp ) printtime() ;
			printf("Error sending abort message\n") ;
			close (controlsock) ;
			return -2 ; /* restart connection and no retry*/
		}
		/*
		** No retry needed
		*/
		return -3 ;
	}
#endif
	/*
	** Now we are ready to create the local file with the correct size
	** and to start the children before sending the real start
	*/
#ifdef STANDART_FILE_CALL
	if ((fd = open(localfile,O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)) < 0 ) {
#else
	if ((fd = open64(localfile,O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)) < 0 ) {
#endif
		if (timestamp ) printtime() ;
		printf("Error creation file %s : %s ",localfile,strerror(errno)) ;

		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		msg = (struct message *)send_buffer ;
		msg->code = MSG_ABR ;
		msg->msglen = 0 ;
		if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			/*
			** Error sending the Abort message
			** Close the connection and ask for a restart of connexion
			*/
			if (timestamp ) printtime() ;
			printf("Error sending abort message\n") ;
			close (controlsock) ;
			return -1 ; /* restart connection */
		}
		/*
		** In future release depending on the open return code we may
		** decide to skip the operation. Now we are going to retry
		*/
		return 2 ;
	}
	unlinkfile = 1 ;
	/*
	** If the file has a zero length do not seek ... Send only a special message
	** saying that everything is OK
	*/
	if ( filelen == 0 ) {
		close(fd) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		msg = (struct message *)send_buffer ;
		msg->code = MSG_CREATE_ZERO ;
		msg->msglen = 0 ;
		if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			/*
			** Error sending the CREATE_ZERO Message
			** Close the connection and ask for a restart of connexion
			*/
			if (timestamp ) printtime() ;
			printf("Error sending create zero message\n") ;
			close (controlsock) ;
			return -1 ; /* restart connection */
		}
		if (timestamp ) printtime() ;
		printf("<< OK\n") ;
		return 0 ;
	}
	/*
	** We lseek the file to reach the good size
	*/
	toseek = filelen - 1 ;
#ifdef STANDART_FILE_CALL
	if ( lseek(fd,toseek,SEEK_SET) < 0 ) {
#else
	if ( lseek64(fd,toseek,SEEK_SET) < 0 ) {
#endif
		/*
		** Now in case of error we have to :
		**	close the file
		**	destroy the file
		**	close all sockets
		**	send an abort message
		*/
		if (timestamp ) printtime() ;
		printf("Error seek file %s : %s ",localfile,strerror(errno)) ;
		close(fd) ;
		unlink(localfile) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		msg = (struct message *)send_buffer ;
		msg->code = MSG_ABR ;
		msg->msglen = 0 ;
		if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			/*
			** Error sending the Abort message
			** Close the connection and ask for a restart of connexion
			*/
			if (timestamp ) printtime() ;
			printf("Error sending abort message\n") ;
			close (controlsock) ;
			return -1 ; /* restart connection */
		}
		/*
		** In future release depending on the open return code we may
		** decide to skip the operation. Now we are going to retry
		*/
		return 2 ;
	}
	/*
	** Write one byte to set the size
	*/
	if ( write(fd,"\0",1) != 1) {
		/*
		** Now in case of error we have to :
		**	close the file
		**	destroy the file
		**	close all sockets
		**	send an abort message
		*/
		if (timestamp ) printtime() ;
		printf("Error writing file %s : %s ",localfile,strerror(errno)) ;
		close(fd) ;
		unlink(localfile) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		msg = (struct message *)send_buffer ;
		msg->code = MSG_ABR ;
		msg->msglen = 0 ;
		if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			/*
			** Error sending the Abort message
			** Close the connection and ask for a restart of connexion
			*/
			if (timestamp ) printtime() ;
			printf("Error sending abort message\n") ;
			close (controlsock) ;
			return -1 ; /* restart connection */
		}
		/*
		** In future release depending on the open return code we may
		** decide to skip the operation. Now we are going to retry
		*/
		return 2 ;
	}
	/*
	** And close the file
	*/
	if ( close(fd) < 0 ) {
		/*
		** Now in case of error we have to :
		**	destroy the file
		**	close all sockets
		**	send an abort message
		*/
		if (timestamp ) printtime() ;
		printf("Error writing file %s : %s ",localfile,strerror(errno)) ;
		unlink(localfile) ;
		/*
		** Close all other sockets
		*/
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		msg = (struct message *)send_buffer ;
		msg->code = MSG_ABR ;
		msg->msglen = 0 ;
		if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			/*
			** Error sending the Abort message
			** Close the connection and ask for a restart of connexion
			*/
			if (timestamp ) printtime() ;
			printf("Error sending abort message\n") ;
			close (controlsock) ;
			return -1 ; /* restart connection */
		}
		/*
		** In future release depending on the open return code we may
		** decide to skip the operation. Now we are going to retry
		*/
		return 2 ;
	}
	/*
	** Everything seems to be ready so start our children
	** but before we are going to calculate the starting position
	** and the offset for each child 
	*/
	nbperchild = (toseek+1)/realnbport ;
	for (i=0 ; i<realnbport ; i++) {
		/*
		** Calculate the start point of the child to be forked 
		** and the number of byte to be transfered
		*/
		if ( i == realnbport-1) {
			startpoint = i * nbperchild ;
			nbtoget = (toseek+1)-(nbperchild*(realnbport-1)) ;
		} else {
			startpoint = i * nbperchild ;
			nbtoget = nbperchild ; 
		}
		if ( (retcode = fork()) == 0 ) {
			/* 
			** In child the first thing to do is to 
			** close the control socket, and all the 
			** socket not concerning itself
			*/
			close(controlsock) ;
			close (fileno(infd)) ;
			/*
			** We only need to close the upper socket
			** because the others have been closed by the
			** father
			*/
			for ( k = i+1 ; k <realnbport ; k++ ) {
				close(datasock[k]) ;
			}
			/*
			** Check if file exist
			*/
#ifdef STANDART_FILE_CALL
			if ( stat(localfile,&statbuf) < 0 ) {
#else
			if ( stat64(localfile,&statbuf) < 0 ) {
#endif
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error stating file %s\n",getpid(),strerror(k)) ;
				_exit(k) ;
			}
			/*
			** open it
			*/
#ifdef STANDART_FILE_CALL
			if ((fd = open(localfile,O_RDWR,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)) < 0 ) {
#else
			if ((fd = open64(localfile,O_RDWR,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)) < 0 ) {
#endif
				k = errno ;
				unlink(localfile) ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error opening file %s : %s\n",getpid(),localfile,strerror(k)) ;
				exit(k) ;
			}
#ifdef STANDART_FILE_CALL
			if ( lseek(fd,startpoint,SEEK_SET) < 0 ) {
#else
			if ( lseek64(fd,startpoint,SEEK_SET) < 0 ) {
#endif
				k = errno ;
				close(fd) ;
				unlink(localfile) ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error seeking file %s : %s\n",getpid(),localfile,strerror(k)) ;
				exit(k)  ;
			}
			/*
			** We are now just going to select on our 
			** socket
			**/
			nfds = sysconf(_SC_OPEN_MAX) ;
			FD_ZERO(&selectmask) ;
			FD_SET(datasock[i],&selectmask) ;
			/*
			** Set the timer for the connection
			*/
			wait_timer.tv_sec  = CHILDWAITTIME ;
			wait_timer.tv_usec = 0 ;
			retcode = select(nfds,&selectmask,0,0,&wait_timer) ;
			if ( retcode < 0 ) {
				/*
				** select return in error 
				** Let us sleep a while in order to let
				** the father fork all its children
				*/
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error select %s\n",getpid(),strerror(k)) ;
				close(fd) ;
				unlink(localfile) ;
				sleep(CHILDWAITTIME) ;
				_exit(k) ;
			}
			if ( retcode == 0 ) {
				/*
				** That is a time out let us exit 
				** with the time out error
				*/
				if (timestamp ) printtime() ;
				printf("Child %06d : Time out on select\n",getpid()) ;
				close(fd) ;
				unlink(localfile) ;
				_exit(ETIMEDOUT) ;
			}
			/*
			** At this point as we only set one bit we have 
			** got a connection
			*/
			if ( (ns = accept(datasock[i],0,0) ) < 0 ) {
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error accept %s\n",getpid(),strerror(errno)) ;
				close(fd) ;
				unlink(localfile) ;
				_exit(k) ;
			}
			/*
			** Close the listening socket
			*/
			close(datasock[i]) ;
			/*
			** Set the receiving socket in non blocking mode
			*/
			if ( fcntl(ns,F_SETFL,O_NONBLOCK) < 0 ) {
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error fcntl socket %s\n",getpid(),strerror(errno)) ;
				_exit(k) ;
			}
			/*
			** Now start the reading loop
			*/
			nbget = 0 ;
			while ( nbget < nbtoget ) {
				/*
				** depending on compresstype the packet format is different
				*/
				if ( compresstype == COMPRESSION ) {
					/*
					** Read the compression header
					*/
					if ( readmessage(ns,receive_buffer,COMPMESSLEN,DATASOCKTO,1) < 0 ) {
						if (timestamp ) printtime() ;
						printf("Child %06d : Error reading compression header\n",getpid()) ;
						close(fd) ;
						unlink(localfile) ;
						_exit(ETIMEDOUT) ;
					}
					msg_compress = (struct mess_compress *) receive_buffer ;
					if ( msg_compress->code == DATA_COMPRESS) {
						compressionon = 1 ;
					} else {
						compressionon = 0 ;
					}
#ifndef WORDS_BIGENDIAN
					datatoreceive = ntohl(msg_compress->datalen) ;
#else
					datatoreceive = msg_compress->datalen ;
#endif
				} else {
					/*
					** No compression 
					*/
					if ( READBUFLEN <= nbtoget-nbget ) {
						datatoreceive = READBUFLEN  ;
					} else {
						datatoreceive = nbtoget-nbget ;
					}
				}
				dataonone = 0 ;
				while ( dataonone < datatoreceive ) {
					nfds = sysconf(_SC_OPEN_MAX) ;
					FD_ZERO(&selectmask) ;
					FD_SET(ns,&selectmask) ;
					wait_timer.tv_sec  = DATASOCKTO ;
					wait_timer.tv_usec = 0 ;
					if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
						/*
						** Select error
						*/
						k = errno ;
						if (timestamp ) printtime() ;
						printf("Child %06d : Error receiving select\n",getpid()) ;
						close(fd) ;
						unlink(localfile) ;
						_exit(k) ;
					} else if ( retcode == 0 ) {
						if (timestamp ) printtime() ;
						printf("Child %06d : Error while receiving : Time Out\n",getpid()) ;
						close(fd) ;
						unlink(localfile) ;
						k=ETIMEDOUT ;
						_exit(k) ;
					} else {
						retcode = recv(ns,&data_buffer[dataonone],datatoreceive-dataonone,0) ;
						if ( retcode < 0 ) {
							if ( errno == EINTR ) {
								if (timestamp ) printtime() ;
								printf("Child %06d : Error while receiving : Interrupted %lld\n",getpid(),nbget) ;
								close(fd) ;
								unlink(localfile) ;
								_exit(EINTR) ;
							} else {
								k = errno ;
								close(fd) ;
								unlink(localfile) ;
								if (timestamp ) printtime() ;
								printf("Child %06d : Error while receiving : %s\n",getpid(),strerror(k)) ;
								_exit(k) ;
							}	
						} else if ( retcode == 0 ) {
							k = ECONNRESET ;
							close(fd) ;
							unlink(localfile) ;
							if (timestamp ) printtime() ;
							printf("Child %06d : Connection breaks : %lld\n",getpid(),nbget) ;
							_exit(k) ;
						} else {
							dataonone = dataonone + retcode ;
						}
					}
				}
				/*
				** Uncompress if necessary
				*/
				if ( compresstype == COMPRESSION ) {
					if ( compressionon == 1 ) {
						bufcomplen = READBUFLEN ;
						buflen = dataonone ;
						retcode = uncompress((Bytef *) comp_buffer, &bufcomplen, (Bytef *) data_buffer, buflen) ;
						if ( retcode != 0 ) {
							if (timestamp ) printtime() ;
							printf("Child %06d :Error while decompressing %d\n",getpid(),retcode) ;
							close(fd) ;
							unlink(localfile) ;
							_exit(EILSEQ) ;
						}
						memcpy(data_buffer,comp_buffer,READBUFLEN) ;
						lentowrite = bufcomplen ;
					} else {
						lentowrite = dataonone ;
					}
				} else {
					lentowrite = dataonone ;
				}
				/*
				** Write to the file
				*/
				lenwrited = 0 ;
				while ( lenwrited < lentowrite ) {
					if ( (retcode = write(fd,&data_buffer[lenwrited],lentowrite-lenwrited)) < 0 ) {
						k = errno ;
						if (timestamp ) printtime() ;
						printf("Child %06d :Error writing file %s\n",getpid(),strerror(k)) ;
						close(fd) ;
						unlink(localfile) ;
						_exit(k)  ;
					}
					lenwrited = lenwrited + retcode ;
				}
				nbget = nbget+lenwrited ;
			}
			/*
			** Close the file
			*/
			if ( close(fd) < 0 ) {
				/*
				** Close failed 
				*/
				k = errno ;
				unlink(localfile) ;
				_exit(k) ;
			}
			/*
			** All the data have correctly arrived so send the ACK
			*/	
			msg = (struct message *) send_buffer ;
			msg->code = MSG_ACK ;
			msg->msglen = 0 ;
			if ( writemessage(ns,send_buffer,MINMESSLEN,DATASOCKTO,1) < 0 ) {
				/*
				** Error sending the ACK message
				*/
				if (timestamp ) printtime() ;
				printf("Error sending ACK message\n") ;
				close(fd) ;
				unlink(localfile) ;
				_exit(ETIMEDOUT) ;
			}
			/*
			** End correct
			*/
			_exit(0) ;
		} else if ( retcode == -1 ) {
			/*
			** We are in the father and the fork has failed
			*/
			if (timestamp ) printtime() ;
			printf("Fork has failed\n") ;
			/*
			** Kill all the already started children
			*/
			for ( k=0 ; k < i ; k++) {
				if (debug) printf("Killing child %d\n",pid_child[k]) ; 
				kill(pid_child[k],SIGKILL) ;
			}
			/*
			** Close all the sockets still open
			*/
			for (k=i ; k<realnbport ; k++) {
				close(datasock[k]) ;
				datasock[k] = 0 ;
			}
			/*
			** Close the control socket and return with
			** code -1
			*/
			close(controlsock) ;
			return -1 ;
		} else {
			/*
			** In father with fork successfull
			*/
			/*
			** Close the socket
			*/
			close (datasock[i]) ;
			datasock[i] = 0 ;
			pid_child[i] = retcode ;
		}
	}
	/*
	** All children are started we are now going to send
	** the start message
	*/
	msg = (struct message *)send_buffer ;
	msg->code = MSG_RETR_START ;
	msg->msglen = 0 ;
	if ( writemessage(controlsock,send_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
		/*
		** Fail to send
		** Close the connection and ask for a restart of connexion
		*/
		if (timestamp ) printtime() ;
		printf("Error sending start retr message\n") ;
		/*
		** Kill all the already started children
		*/
		for ( i=0 ; i < realnbport ; i++) {
			if ( pid_child[i] != 0) {
				if (debug) printf("Killing child %d\n",pid_child[k]) ; 
				kill(pid_child[k],SIGKILL) ;
			}
		}
		unlink (localfile) ;
		close (controlsock) ;
		return -1 ;
	}
	/*
	** now wait for the end of the game
	*/
	(void) gettimeofday(&tstart, (struct timezone *)0);
	k = 0 ;
waitcontrol:
	nfds = sysconf(_SC_OPEN_MAX) ;
	FD_ZERO(&selectmask) ;
	FD_SET(controlsock,&selectmask) ;
	wait_timer.tv_sec  = CHECKCHILDTIME ;
	wait_timer.tv_usec = 0 ;
	retcode = select(nfds,&selectmask,0,0,&wait_timer) ;
	if ( retcode < 0 ) {
		/*
		** Select error
		*/
		if ( errno != EINTR ) {
			/*
			** we have got an error so close the connection
			** and restart
			*/
			if (timestamp ) printtime() ;
			printf("Error select waiting on control connection: %s\n ",strerror(errno) ) ;
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			unlink(localfile) ;
			close(controlsock) ;
			return -1 ;
		} else {
			/*
			** Interrupted by a signal
			*/
			goto waitcontrol ;
		}
	} else if ( retcode == 0 ) {
		/*
		** Check if all child are not dead
		*/
		for ( i = 0 ; i<realnbport; i++) {
			if ( pid_child[i] != 0 ) {
				pid = waitpid(pid_child[i],&status,WNOHANG) ;
				if ( pid < 0 ) {
					if ( errno != EINTR ) {
						if ( errno == ECHILD ) {
							/*
							** Chid has terminated
							*/
							pid_child[i] = 0 ;
						}
					}
				}
			}
		}
		for ( i = 0 ; i<realnbport; i++) {
			if ( pid_child[i] != 0 ) {
				/*
				** There is still a child so continue waiting on 
				** control socket
				*/
				if ( debug ) printf("Still : %d\n",pid_child[i]) ;
				goto waitcontrol ;
			}
		}
		if ( k == 0 ) {
			/*
			** First time we get here let a chance
			** to get an answer on control socket
			*/
			k = 1 ; 
			goto waitcontrol ;
		}
		/*
		** Second time without any child 
		** Bad thing so abort 
		*/
		if (timestamp ) printtime() ;
		printf("No more children in waitcontrol\n") ;
		unlink(localfile) ;
		close(controlsock) ;
		return -1 ;
	} else {
		/*
		** read the message
		*/
		if ( readmessage(controlsock,receive_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			if (timestamp ) printtime() ;
			printf("Error reading control message after waitcontrol\n") ;
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			close(controlsock) ; 
			unlink(localfile) ;
			return -1 ;
		}
		msg = (struct message *)receive_buffer ;
		if ( msg->code == MSG_BAD ) {
			/*
			** The transfer has failed so ...
			*/
#ifndef WORDS_BIGENDIAN
			msglen = ntohl(msg->msglen) ;
#else
			msglen = msg->msglen ;
#endif
			if ( readmessage(controlsock,receive_buffer,msglen,CONTROLSOCKTO,0) < 0 ) {
				if (timestamp ) printtime() ;
				printf("Error reading BAD message after send waitcontrol\n") ;
				/*
				** Kill all child 
				*/
				for ( i = 0 ; i<realnbport; i++) {
					if ( pid_child[i] != 0 ) {
						if (debug) printf("Killing child %d\n",pid_child[i]) ; 
						kill(pid_child[i],SIGKILL) ;
						pid_child[i] = 0 ;
					}
				}
				close(controlsock) ;
				unlink(localfile) ;
				return -1 ;
			} else {
				/*
				** Kill all child 
				*/
				for ( i = 0 ; i<realnbport; i++) {
					if ( pid_child[i] != 0 ) {
						if (debug) printf("Killing child %d\n",pid_child[i]) ; 
						kill(pid_child[i],SIGKILL) ;
						pid_child[i] = 0 ;
					}
				}
				receive_buffer[msglen] = '\0' ;
				if (timestamp ) printtime() ;
				printf("<< %s\n",receive_buffer) ;
				unlink(localfile) ;
				return 1;
			}
		} else if (msg->code == MSG_OK ) {
			/*
			** At this stage the transfer is OK so we do not
			** care about the message, so return 0 and let see
			** on next send if we got an error. Probably the
			** message after OK will be soon suppressed.
			*/
#ifndef WORDS_BIGENDIAN
			msglen = ntohl(msg->msglen) ;
#else
			msglen = msg->msglen ;
#endif
			readmessage(controlsock,receive_buffer,msglen,CONTROLSOCKTO,0) ;
			if (timestamp ) printtime() ;
			printf("<< OK\n") ;
			(void) gettimeofday(&tend, (struct timezone *)0);
			tdiff.tv_sec = tend.tv_sec - tstart.tv_sec ;
			tdiff.tv_usec = tend.tv_usec - tstart.tv_usec;
			if (tdiff.tv_usec < 0) tdiff.tv_sec--, tdiff.tv_usec += 1000000;
			s = tdiff.tv_sec + (tdiff.tv_usec / 1000000.);
#define nz(x)   ((x) == 0 ? 1 : (x))
			bs = filelen / nz(s);
			if (timestamp ) printtime() ;
			printf("%" LONG_LONG_FORMAT" bytes got in %.3g secs (%.3g Kbytes/sec or %.3g Mbits/s)\n", filelen, s, bs / 1024.0,(8.0*bs) / (1024.0 * 1024.0));
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			return 0 ;
		} else {
			/*
			** Receive unkwown message so something is
			** going wrong. close the control socket
			** and restart
			*/
			if (timestamp ) printtime() ;
			printf("Receive unkown message waitcontrol : %d\n",msg->code) ;
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			close(controlsock) ; 
			unlink(localfile) ;
			return -1 ;
		}
	}
}
