/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN :
 	 BB_RET_OK                  = everything OK 
	 BB_RET_FT_NR               = Fatal error no retry needed on this command but connection has been
	                              kept
	 BB_RET_ERROR               = Error but retry possible
	   
 mputfile.c v 1.8.8 2000/06/02	- Creation of the routine 
 		 	v 1.8.9 2000/08/08	- Add timestamp
 		 	v 1.9.0 2000/08/08	- Add -r option

*****************************************************************************/
#include <glob.h>
#include <stdio.h>

#include <client.h>

extern int	timestamp ;
extern int	globaltrymax;

int mputfile(char *line, int *controlsock, int nbport, int compresstype, FILE *resfd,char *username, char *password) 
{
	int	nbitem ; /* Number of item in line, must be equal to 2 */
	char	action[MAXLINE] ;
	char	localfile[MAXLINE] ;
	glob_t pg ;
	int retcode ;
	int	i ;
	char	*tmpfile ;
	char	newline[MAXLINE] ;
	int		tmpsock ;
	int 	connectionisbroken ;
	int		nbtry ;

	nbitem = sscanf(line,"%s %s",action,localfile) ;
	
	if (nbitem != 2 ) {
		if (timestamp ) printtime() ;
		printf("Incorrect mput command : %s\n",line) ;
		return BB_RET_FT_NR ;
	}
	if (timestamp ) printtime() ;
	printf(">> COMMAND : %s\n",line) ;
	/*
	** We are going to get all files matching
	*/
	if ( (retcode = glob(localfile,0,0,&pg)) < 0 ) {
		if ( retcode == GLOB_NOMATCH ) {
			/*
			** No files were corresponding
			*/
			if (timestamp ) printtime() ;
			printf("<< OK : No file to transfer\n") ;
			globfree(&pg) ;
			return BB_RET_OK ;
		} else {
			if (timestamp ) printtime() ;
			printf("Error during filelist completion\n") ;
			globfree(&pg) ;
			return BB_RET_ERROR ;
		}
	}
	/*
	** Now make a loop on each file
	*/
	connectionisbroken = 0 ;
	for (i = 0 ; i < pg.gl_pathc ; i++ ) {
		tmpfile = pg.gl_pathv[i] ;
		sprintf(newline,"put %s",tmpfile) ;
		tmpsock = *controlsock ;
		for ( nbtry = 1 ; nbtry <= globaltrymax ; nbtry++ ) {
			/*
			** check if the last try was not disconnected
			*/
			if ( connectionisbroken == 1 ) {
				reconnecttoserver(username,password,controlsock) ;
				connectionisbroken = 0 ;
			}
			tmpsock = *controlsock ;
			retcode = putafile(newline,tmpsock,nbport,compresstype) ;
			if ( retcode == BB_RET_FT_NR ) {
				/*
				** Fatal error no retry and connection is not broken
				*/
				break ;
			} else if ( retcode == BB_RET_FT_NR_CONN_BROKEN ) {
				/*
				** Fatal error no retry and connection is broken
				*/
				connectionisbroken = 1 ;
				break ;
			} else if ( retcode == BB_RET_OK ) {
				/*
				** Success , no retry
				*/							
				break ;
			} else if ( retcode == BB_RET_CONN_BROKEN ) {
				/*
				** Retry needed with a new connection
				*/
				if ( nbtry != globaltrymax ) {
					if (timestamp ) printtime() ;
					printf("Retrying connection waiting %d s\n",WAITRETRYTIME) ;
					sleep(WAITRETRYTIME) ;
					reconnecttoserver(username,password,controlsock) ;
				} else {
					/*
					** In case of broken connection and max retry reached
					** indicate to the next transfer that the connection
					** is broken
					*/
					connectionisbroken = 1 ;
				}
			} else {
				/*
				** retcode > 0 means retry on this transfer
				*/
				if ( nbtry != globaltrymax ) {
					if (timestamp ) printtime() ;
					printf("Retrying transfer waiting %d s\n",WAITRETRYTIME) ;
					sleep(WAITRETRYTIME) ;
				}
			}
		}
		if ( retcode != 0 && nbtry == globaltrymax+1) {
			if (timestamp ) printtime() ;
			printf("Maximum try reached (%d) skipping file\n",globaltrymax) ;
			fprintf(resfd,"%s FAILED\n",newline) ;
			fflush(resfd) ;
		} else if ( retcode == 0 ) {
			fprintf(resfd,"%s OK\n",newline) ;
			fflush(resfd) ;
		} else {
			fprintf(resfd,"%s FAILED\n",newline) ;
			fflush(resfd) ;
		}
	}
	globfree(&pg) ;
	return BB_RET_OK ;
}
