/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN :
 	 0    = everything OK file correctly transfered
	 -1   = connection with server broken retry with connection first
	 -2   = Fatal error no retry needed on this command and connection is
	        broken
	 <-2  = Fatal error no retry needed on this command but connection has been
	        kept
	 >0   = Error but retry possible
	   
 putafile.c v 1.5.0  2000/03/23	- Creation of the routine in order to replace
                                  sendfile.c and sendfilecompress.c
 			v 1.7.1  2000/04/04	- Correct printf on error (detected on SGI)
			v 1.8.0  2000/04/14	- Change location of headers file for zlib
			v 1.8.3  2000/04/18	- Implement the MSG_BAD_NO_RETRY message and
								  change the return code (-2 added)
			v 1.8.7  2000/05/28	- Portage to OSF1
				   				- Set socket in non blocking mode
								- Change headers
			v 1.8.8  2000/05/30	- Check if file to send is not a directory
								- Accept put command with one filename
								- Take care of zero length file
 		 	v 1.8.9  2000/08/08	- Add timestamp
								- Add Child check in waitcontrol
		    v 1.8.10 2000/08/10	- Portage to Linux
			v 1.9.0  2000/08/21	- Use configure to help portage

*****************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <bbftp.h>
#include <client.h>
#include <common.h>
#include <structures.h>

#include <zlib/zlib.h>

#ifndef WORDS_BIGENDIAN
my64_t	convertlong(my64_t v) ;
#endif

extern struct sockaddr_in data_source[MAXPORT] ;
extern struct sockaddr_in hisctladdr ;
extern struct sockaddr_in myctladdr ;
extern int		datasock[MAXPORT] ; 
extern int		pid_child[MAXPORT] ;
extern int		debug ;
extern int	timestamp ;
extern	FILE	*infd ;

int putafile(char *line, int controlsock, int nbport, int compresstype) 
{
	int	nbitem ; /* Number of item in line, must be equal to 2 or 3 */
	char	action[MAXLINE] ;
	char	localfile[MAXLINE] ;
	char	remotefile[MAXLINE] ;
	
 	int		fd ;
	my64_t	filelen ;
	int		realnbport ;
	
#ifdef STANDART_FILE_CALL
	off_t	tmpnbport ;
	off_t	startpoint ;
	off_t	nbperchild ;
	off_t 	nbtosend ;
	off_t 	nbread ;
	off_t 	nbsent ;
	off_t 	numberread ;
	off_t	realnbtosend ;
	off_t	toseek ;
	struct stat statbuf ;
#else
	off64_t		tmpnbport ;
	off64_t		startpoint ;
	off64_t		nbperchild ;
	off64_t 	nbtosend ;
	off64_t 	nbread ;
	off64_t 	nbsent ;
	off64_t 	numberread ;
	off64_t		realnbtosend ;
	off64_t		toseek ;
	struct stat64 statbuf	;
#endif
	
	
	int		lentosend ;
	
	fd_set	selectmask ; /* Select mask */
	int	nfds ; /* Max number of file descriptor */
	struct timeval	wait_timer;
	int	ns ;
	int	i ,k;
	int retcode ,msglen;
	int		code ;
	int		status ;
	pid_t	pid ;
	char	readbuffer[READBUFLEN] ;
	char	buffercomp[READBUFLEN] ;
	struct	mess_compress *msg_compress ;
	char	compress_mess[COMPMESSLEN] ;
	uLong	buflen ;
	uLong	bufcomplen ;
	
	char	send_buffer[MAXMESSLEN] ;
	char	answer_buffer[MAXMESSLEN] ;
 	struct	message *msg ;
	struct	mess_store *msg_store ;

	struct timeval  tstart, tend ,tdiff;
	float	s, bs;
	int		savederrno ;
	
	/*
	** Initialize the pid array (for the case of SIGTERM)
	*/
	for ( i=0 ; i< MAXPORT ; i++) {
		pid_child[i] = 0 ;
	}
	nbitem = sscanf(line,"%s %s %s",action,localfile,remotefile) ;
	
	if (nbitem == 2 ) {
		/*
		** This mean that remotefile = localfile
		*/
		strcpy(remotefile,localfile) ;
	} else if (nbitem != 3 ) {
		if (timestamp ) printtime() ;
		printf("Incorrect put command : %s\n",line) ;
		return -3 ;
	}
	if (timestamp ) printtime() ;
	printf(">> COMMAND : %s\n",line) ;
	/*
	** Before doing anything we are going to check if the local
	** file is not a directory
	*/
#ifdef STANDART_FILE_CALL
	if ( stat(localfile,&statbuf) < 0 ) {
#else
	if ( stat64(localfile,&statbuf) < 0 ) {
#endif
		/*
		** We will not retry in the following case  :
		**		EACCES		: Search permission denied
		**		ELOOP		: To many symbolic links on path
		**		ENAMETOOLONG: Path argument too long
		**		ENOENT		: The file does not exists
		**		ENOTDIR		: A component in path is not a directory
		*/
		savederrno = errno ;
		if (timestamp ) printtime() ;
		printf("Error stating file %s : %s \n",localfile,strerror(savederrno)) ;
		if ( savederrno == EACCES ||
			savederrno == ELOOP ||
			savederrno == ENAMETOOLONG ||
			savederrno == ENOENT ||
			savederrno == ENOTDIR ) {
			return -3 ;
		} else {
			return 1 ;
		}
	} else {
		/*
		** The file exists so check if it is a directory
		*/
		if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
			if (timestamp ) printtime() ;
			printf("File %s is a directory\n",localfile) ;
			return -3 ;
		}
	}				
	/*
	** Clear the data_source structure
	*/
	for (i=0 ; i< MAXPORT ; i++) {
		data_source[i].sin_family = myctladdr.sin_family ;
		data_source[i].sin_addr.s_addr = myctladdr.sin_addr.s_addr ;
		data_source[i].sin_port = 0;
		datasock[i] = 0 ;
		pid_child[i] = 0 ;
	}
	/*
	** We are going to get the file length in order to know if
	** it is useful to use parallel streams. The rule will be
	** if the file length is less than the send buffer size 
	** we only use one stream
	*/
#ifdef STANDART_FILE_CALL
	if ( (fd = open(localfile,O_RDONLY)) < 0 ) {
#else
	if ( (fd = open64(localfile,O_RDONLY)) < 0 ) {
#endif
		/*
		** An error on openning the local file is considered
		** as fatal. Maybe this need to be improved depending
		** on errno
		*/
		if (timestamp ) printtime() ;
		printf("Error opening local file %s : %s\n",localfile,strerror(errno)) ;
		return -3 ;
	}
#ifdef STANDART_FILE_CALL
	if ( (toseek = lseek(fd,0,SEEK_END) ) < 0 ) {
#else
	if ( (toseek = lseek64(fd,0,SEEK_END) ) < 0 ) {
#endif
		/*
		** An error on seekin the local file is considered
		** as fatal. lseek error  in this case is completly
		** abnormal
		*/
		close(fd) ;
		if (timestamp ) printtime() ;
		printf("Error seeking local file %s : %s\n",localfile,strerror(errno)) ;
		return -3 ;
	}
	filelen = toseek ;
	/*
	** Close the file as the only interresting thing was the length
	*/
	close(fd) ;
	/*
	** If file length is zero just set realnbport to 0
	*/
	if ( filelen == 0 ) {
		realnbport = 0 ;
	} else {
		/*
		** Calculate the useful number of port
		*/
		tmpnbport = toseek / TCPWINSIZE ;
		if ( tmpnbport == 0 ) {
			realnbport = 1 ;
		} else if ( tmpnbport < nbport ) {
			realnbport = tmpnbport ;
		} else {
			realnbport = nbport ;
		}
	}
	/*
	** Get the data socket
	*/
	if ( getdatasock(realnbport) < 0 ) {
		if (timestamp ) printtime() ;
		printf("Error getting data socket\n") ;
		/*
		** In case of error getdatasock does not clean the
		** structure and does not close the sockets, it is
		** the calling program's work
		*/ 
		for (i=0 ; i<realnbport ; i++) {
			if ( datasock[i] != 0 ) {
				close(datasock[i]) ;
				datasock[i] = 0 ;
			}
		}
		/*
		** In this case we allow retry
		*/
		return 1 ;
	}
	/*
	** Before sending the control message we are going to fork 
	** all the childs and to keep in mind their pid
	*/
	if ( filelen == 0 ) {
		nbperchild = 0 ;
	} else {
		nbperchild = toseek/realnbport ;
	}
	for (i=0 ; i<realnbport ; i++) {
		/*
		** Calculate the start point of the child to be forked 
		** and the number of byte to be transfered
		*/
		if ( i == realnbport-1) {
			startpoint = i * nbperchild ;
			nbtosend = toseek -(nbperchild*(realnbport-1)) ;
		} else {
			startpoint = i * nbperchild ;
			nbtosend = nbperchild ; 
		}
		if ( (retcode = fork()) == 0 ) {
			/* 
			** In child the first thing to do is to 
			** close the control socket, and all the 
			** socket not concerning itself
			*/
			close(controlsock) ;
			close (fileno(infd)) ;
			/*
			** We only need to close the upper socket
			** because the others have been closed by the
			** father
			*/
			for ( k = i+1 ; k <realnbport ; k++ ) {
				close(datasock[k]) ;
			}
			/*
			** We are now just going to select on our 
			** socket
			**/
			nfds = sysconf(_SC_OPEN_MAX) ;
			FD_ZERO(&selectmask) ;
			FD_SET(datasock[i],&selectmask) ;
			/*
			** Set the timer for the connection
			*/
			wait_timer.tv_sec  = CHILDWAITTIME ;
			wait_timer.tv_usec = 0 ;
			retcode = select(nfds,&selectmask,0,0,&wait_timer) ;
			if ( retcode < 0 ) {
				/*
				** select return in error 
				** Let us sleep a while in order to let
				** the father fork all its children
				*/
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error select %s\n",getpid(),strerror(k)) ;
				sleep(CHILDWAITTIME) ;
				_exit(k) ;
			}
			if ( retcode == 0 ) {
				/*
				** That is a time out let us exit 
				** with the time out error
				*/
				if (timestamp ) printtime() ;
				printf("Child %06d : Time out on select\n",getpid()) ;
				_exit(ETIMEDOUT) ;
			}
			/*
			** At this point as we only set one bit in mask so we have 
			** got a connection
			*/
			if ( (ns = accept(datasock[i],0,0) ) < 0 ) {
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error accept %s\n",getpid(),strerror(errno)) ;
				_exit(k) ;
			}
			/*
			** Close the listening socket
			*/
			close(datasock[i]) ;
			/*
			** Set the receiving socket in non blocking mode
			*/
			if ( fcntl(ns,F_SETFL,O_NONBLOCK) < 0 ) {
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error fcntl socket %s\n",getpid(),strerror(errno)) ;
				_exit(k) ;
			}
			/*
			** Open the file and seek to the good position
			*/
#ifdef STANDART_FILE_CALL
			if ( ( fd = open(localfile,O_RDONLY) ) < 0 ) {
#else
			if ( ( fd = open64(localfile,O_RDONLY) ) < 0 ) {
#endif
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error opening file %s\n",getpid(),localfile) ;
				_exit(k) ;
			}
#ifdef STANDART_FILE_CALL
			if ( lseek(fd,startpoint,SEEK_SET) < 0 ) {
#else
			if ( lseek64(fd,startpoint,SEEK_SET) < 0 ) {
#endif
				k = errno ;
				if (timestamp ) printtime() ;
				printf("Child %06d : Error seeking file %s\n",getpid(),localfile) ;
				_exit(k) ;
			}
			/*
			** Now start the reading loop
			*/
			nbread = 0 ;
			if (debug) printf("Child %06d : port %d : starting point %lld ,number to send %lld \n",getpid(),data_source[i].sin_port,startpoint,nbtosend) ;
			while ( nbread < nbtosend ) {
				if ( (numberread = read ( fd, readbuffer, (sizeof(readbuffer) <= nbtosend - nbread) ? sizeof(readbuffer) : nbtosend-nbread) ) > 0 ) {
					if (debug) printf("Child %06d : numberread %lld\n",getpid(),numberread) ;
					nbread = nbread+numberread ;
					if ( compresstype == COMPRESSION ) {
						/*
						** In case of compression we are going to use
						** a temporary buffer
						*/
						bufcomplen = READBUFLEN ;
						buflen = numberread ;
						msg_compress = ( struct mess_compress *) compress_mess ;
						retcode = compress((Bytef *)buffercomp,&bufcomplen,(Bytef *)readbuffer,buflen) ;
						if ( retcode != 0 ) {
							/*
							** Compress error, in this cas we are sending the
							** date uncompressed
							*/
							msg_compress->code = DATA_NOCOMPRESS ;
#ifndef WORDS_BIGENDIAN
							msg_compress->datalen = ntohl(numberread) ;
#else
							msg_compress->datalen = numberread ;
#endif
							realnbtosend = numberread ;
							if ( debug) printf("Child %06d : Compress failed %d\n",getpid(),retcode) ;
						} else {
							msg_compress->code = DATA_COMPRESS ;
#ifndef WORDS_BIGENDIAN
							msg_compress->datalen = ntohl(bufcomplen) ;
#else
							msg_compress->datalen = bufcomplen ;
#endif
							realnbtosend =  bufcomplen ;
							memcpy(readbuffer,buffercomp,READBUFLEN) ;
							if ( debug) printf("Child %06d : Compress successful newlen :%ld\n",getpid(),bufcomplen) ;
						}
						/*
						** Send the header
						*/
						if ( writemessage(ns,compress_mess,COMPMESSLEN,DATASOCKTO,1) < 0) {
							k = ETIMEDOUT ;
							if (timestamp ) printtime() ;
							printf("Error sending header data \n") ;
							_exit(k) ;
						}
					} else {
						realnbtosend = numberread ;
					}
					/*
					** Send the data
					*/
					nbsent = 0 ;
					while ( nbsent < realnbtosend ) {
						lentosend = realnbtosend-nbsent ;
						nfds = sysconf(_SC_OPEN_MAX) ;
						FD_ZERO(&selectmask) ;
						FD_SET(ns,&selectmask) ;
						wait_timer.tv_sec  = DATASOCKTO  ;
						wait_timer.tv_usec = 0 ;
						if ( (retcode = select(nfds,0,&selectmask,0,&wait_timer) ) == -1 ) {
							/*
							** Select error
							*/
							k = errno ;
							if (timestamp ) printtime() ;
							printf("Child %06d : Error sending select : %s\n",getpid(),strerror(errno)) ;
							_exit(k) ;
						} else if ( retcode == 0 ) {
							k=ETIMEDOUT ;
							if (timestamp ) printtime() ;
							printf("Child %06d : Error sending Time Out\n",getpid()) ;
							_exit(k) ;
						} else {
							retcode = send(ns,&readbuffer[nbsent],lentosend,0) ;
							if ( retcode < 0 ) {
								k = errno ;
								if (timestamp ) printtime() ;
								printf("Child %06d : Error sending retcode %d : %s\n",getpid(),retcode,strerror(errno)) ;
								_exit(k) ;
							} else if ( retcode == 0 ) {
								k = ECONNRESET ;
								if (timestamp ) printtime() ;
								printf("Child %06d : Connexion breaks\n",getpid()) ;
								_exit(k) ;
							} else {
								nbsent = nbsent+retcode ;
							}
						}
					}
				} else {
					k = errno ;
					if (timestamp ) printtime() ;
					printf("Child %06d : Error reading : %s",getpid(),strerror(errno)) ;
					_exit(k) ;
				}
			}
			/*
			** All data has been sent so wait for the acknoledge
			*/
			if (debug) printf("Child %06d : Waiting ACK port number %d\n",getpid(),data_source[i].sin_port) ;
						
			if ( readmessage(ns,answer_buffer,MINMESSLEN,ACKTO,1) < 0 ) {
				if (timestamp ) printtime() ;
				printf("Child %06d : Error reading ACK \n",getpid()) ;
				_exit(1) ;
			}
			msg = (struct message *)answer_buffer ;
			if ( msg->code == MSG_ACK ) {
				if (debug) printf("Child %06d : End correct\n",getpid()) ;
				close(ns) ;
			} else {
				if (timestamp ) printtime() ;
				printf("Child %06d : Receive unkown message %d\n",getpid(),msg->code) ;
				close(ns) ;
				_exit(1) ;
			}
			_exit(0) ;
		} else if ( retcode == -1 ) {
			/*
			** We are in the father and the fork has failed
			*/
			if (timestamp ) printtime() ;
			printf("Fork has failed\n") ;
			/*
			** Kill all the already started children
			*/
			for ( k=0 ; k < i ; k++) {
				if (debug) printf("Killing child %d\n",pid_child[k]) ; 
				kill(pid_child[k],SIGKILL) ;
			}
			/*
			** Close all the sockets still open
			*/
			for (k=i ; k<realnbport ; k++) {
				close(datasock[k]) ;
				datasock[k] = 0 ;
			}
			/*
			** Close the control socket and return with
			** code -1
			*/
			close(controlsock) ;
			return -1 ;
		} else {
			/*
			** In father with fork successfull
			*/
			/*
			** Close the socket
			*/
			close (datasock[i]) ;
			datasock[i] = 0 ;
			pid_child[i] = retcode ;
		}
	}
	/*
	** Now send the control message 
	*/
	/* 
	** First part
	*/
	msg = (struct message *)send_buffer ;
	if ( compresstype == NOCOMPRESSION ) {
		msg->code = MSG_STORE ;
	} else {
		msg->code = MSG_STORE_C ;
	}
#ifndef WORDS_BIGENDIAN
	msg->msglen = ntohl(STORMESSLEN) ;
#else
	msg->msglen = STORMESSLEN ;
#endif
	if ( writemessage(controlsock,send_buffer,MINMESSLEN,SENDCONTROLTO,0) < 0 ) {
		/*
		** We were not able to send the minimum message so
		** we are going to close the control socket and to 
		** tell the calling program to restart a connection
		*/
		/*
		** First kill all child
		*/
		if (timestamp ) printtime() ;
		printf("Error sending store message part 1 : %s\n",strerror(errno)) ;
		for ( i=0 ; i<realnbport ; i++) {
			if (debug) printf("Killing child %d\n",pid_child[i]) ; 
			kill(pid_child[i],SIGKILL) ;
			pid_child[i] = 0 ;
		}
		close(controlsock) ;
		return -1 ; /* restart connection */
	}
	/* 
	** Second part
	*/
	msg_store = (struct mess_store *)send_buffer ;
	strcpy(msg_store->filename,remotefile)  ;
#ifndef WORDS_BIGENDIAN
	msg_store->filesize = convertlong(filelen) ;
	msg_store->nbport = ntohl(realnbport) ;
#else
	msg_store->filesize = filelen ;
	msg_store->nbport = realnbport ;
#endif
	for (i=0 ; i< realnbport ; i++) {
#ifndef WORDS_BIGENDIAN
		msg_store->port[i]  = ntohl(ntohs(data_source[i].sin_port)) ;
#else
		msg_store->port[i]  = data_source[i].sin_port ;
#endif
		if ( debug ) printf("Using Port number = %d\n", data_source[i].sin_port) ;
	}
	if ( writemessage(controlsock,send_buffer,STORMESSLEN,SENDCONTROLTO,0) < 0) {
		if (timestamp ) printtime() ;
		printf("Error sending store message part 2 : %s\n",strerror(errno)) ;
		for ( i=0 ; i<realnbport ; i++) {
			if (debug) printf("Killing child %d\n",pid_child[i]) ; 
			kill(pid_child[i],SIGKILL) ;
			pid_child[i] = 0 ;
		}
		close(controlsock) ;
		return -1 ; /* restart connection */
	}
	/*
	** Now we are going to wait for the message on the control 
	** connection
	*/
	(void) gettimeofday(&tstart, (struct timezone *)0);
	k = 0 ;
waitcontrol:
	nfds = sysconf(_SC_OPEN_MAX) ;
	FD_ZERO(&selectmask) ;
	FD_SET(controlsock,&selectmask) ;
	wait_timer.tv_sec  = CHECKCHILDTIME ;
	wait_timer.tv_usec = 0 ;
	retcode = select(nfds,&selectmask,0,0,&wait_timer) ;
	if ( retcode < 0 ) {
		/*
		** Select error
		*/
		if ( errno != EINTR ) {
			/*
			** we have got an error so close the connection
			** and restart
			*/
			if (timestamp ) printtime() ;
			printf("Error select waiting on control connection: %s\n ",strerror(errno) ) ;
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			close(controlsock) ;
			return -1 ;
		} else {
			/*
			** Interrupted by a signal
			*/
			FD_ZERO(&selectmask) ;
			FD_SET(controlsock,&selectmask) ;
			goto waitcontrol ;
		}
	} else if ( retcode == 0 ) {
		/*
		** Check if all child are not dead
		*/
		for ( i = 0 ; i<realnbport; i++) {
			if ( pid_child[i] != 0 ) {
				pid = waitpid(pid_child[i],&status,WNOHANG) ;
				if ( pid < 0 ) {
					if ( errno != EINTR ) {
						if ( errno == ECHILD ) {
							/*
							** Chid has terminated
							*/
							pid_child[i] = 0 ;
						}
					}
				}
			}
		}
		for ( i = 0 ; i<realnbport; i++) {
			if ( pid_child[i] != 0 ) {
				/*
				** There is still a child so continue waiting on 
				** control socket
				*/
				if ( debug ) printf("Still : %d\n",pid_child[i]) ;
				goto waitcontrol ;
			}
		}
		if ( k == 0 ) {
			/*
			** First time we get here let a chance
			** to get an answer on control socket
			*/
			k = 1 ; 
			goto waitcontrol ;
		}
		/*
		** Second time without any child 
		** Bad thing so abort 
		*/
		if (timestamp ) printtime() ;
		printf("No more children in waitcontrol\n") ;
		close(controlsock) ;
		return -1 ;
	} else {
		/*
		** read the message
		*/
		if ( readmessage(controlsock,answer_buffer,MINMESSLEN,CONTROLSOCKTO,0) < 0 ) {
			if (timestamp ) printtime() ;
			printf("Error reading control message after waitcontrol\n") ;
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			close(controlsock) ; 
			return -1 ;
		}
		msg = (struct message *)answer_buffer ;
		code = msg->code ;
		if ( code == MSG_BAD || code == MSG_BAD_NO_RETRY) {
			/*
			** The transfer has failed so ...
			*/
#ifndef WORDS_BIGENDIAN
			msglen = ntohl(msg->msglen) ;
#else
			msglen = msg->msglen ;
#endif
			if ( readmessage(controlsock,answer_buffer,msglen,CONTROLSOCKTO,0) < 0 ) {
				if (timestamp ) printtime() ;
				printf("Error reading BAD message after send waitcontrol\n") ;
				/*
				** Kill all child 
				*/
				for ( i = 0 ; i<realnbport; i++) {
					if ( pid_child[i] != 0 ) {
						if (debug) printf("Killing child %d\n",pid_child[i]) ; 
						kill(pid_child[i],SIGKILL) ;
						pid_child[i] = 0 ;
					}
				}
				close(controlsock) ;
				if ( code == MSG_BAD ) {
					return -1 ;
				} else {
					return -2 ;
				}
			} else {
				/*
				** Kill all child 
				*/
				for ( i = 0 ; i<realnbport; i++) {
					if ( pid_child[i] != 0 ) {
						if (debug) printf("Killing child %d\n",pid_child[i]) ; 
						kill(pid_child[i],SIGKILL) ;
						pid_child[i] = 0 ;
					}
				}
				answer_buffer[msglen] = '\0' ;
				if (timestamp ) printtime() ;
				printf("<< %s\n",answer_buffer) ;
				if ( code == MSG_BAD ) {
					return 1;
				} else {
					/*
					** In case of no retry we are going to wait 2 seconds
					** in order for the server to finish its cleaning (in
					** fact if the error message is du to a server child
					** abnormal termination, the message is sent on the death
					** of the first child and the server has some cleaning
					** to do).
					*/
					sleep(1) ;
					return -3 ;
				}
			}
		} else if (msg->code == MSG_OK ) {
			/*
			** At this stage the transfer is OK so we do not
			** care about the message, so return 0 and let see
			** on next send if we got an error. Probably the
			** message after OK will be soon suppressed.
			*/
#ifndef WORDS_BIGENDIAN
			msglen = ntohl(msg->msglen) ;
#else
			msglen = msg->msglen ;
#endif
			readmessage(controlsock,answer_buffer,msglen,CONTROLSOCKTO,0) ;
			if (timestamp ) printtime() ;
			printf("<< OK\n") ;
			(void) gettimeofday(&tend, (struct timezone *)0);
			tdiff.tv_sec = tend.tv_sec - tstart.tv_sec ;
			tdiff.tv_usec = tend.tv_usec - tstart.tv_usec;
			if (tdiff.tv_usec < 0) tdiff.tv_sec--, tdiff.tv_usec += 1000000;
			s = tdiff.tv_sec + (tdiff.tv_usec / 1000000.);
#define nz(x)   ((x) == 0 ? 1 : (x))
			bs = filelen / nz(s);
			if (timestamp ) printtime() ;
			printf("%" LONG_LONG_FORMAT " bytes sent in %.3g secs (%.3g Kbytes/sec or %.3g Mbits/s)\n", filelen, s, bs / 1024.0,(8.0*bs) / (1024.0 * 1024.0));
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			return 0 ;
		} else {
			/*
			** Receive unkwown message so something is
			** going wrong. close the control socket
			** and restart
			*/
			if (timestamp ) printtime() ;
			printf("Receive unkown message waitcontrol : %d\n",msg->code) ;
			/*
			** Kill all child 
			*/
			for ( i = 0 ; i<realnbport; i++) {
				if ( pid_child[i] != 0 ) {
					if (debug) printf("Killing child %d\n",pid_child[i]) ; 
					kill(pid_child[i],SIGKILL) ;
					pid_child[i] = 0 ;
				}
			}
			close(controlsock) ; 
			return -1 ;
		}
	}
}
