/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 bbftpd.c v 0.0.0  1999/11/24
          v 1.3.0  2000/03/16	- Version in openlog
		  v 1.4.0  2000/03/22	- Modify the connection protocole
		                          in order to send the crypt type
								  after the connection.
		  v 1.5.0  2000/03/24	- Change the wait timer in default
		  v 1.6.0  2000/03/24	- Change the main loop to make it more
		  						  modular
		  v 1.6.1  2000/03/24	- Portage on OSF1
		  						- Make the control socket non blocking
		  v 1.8.0  2000/04/14	- Introduce RSA Cryptage
		  v 1.8.6  2000/05/21	- Allow to run with inetd
		  v 1.8.7  2000/05/24	- Introduce version.h and config.h
		  v 1.8.10 2000/08/11	- Portage to Linux
		  v 1.9.0  2000/08/18	- Use configure to help portage
								- default time out set to 900 s

*****************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>

#include <bbftpd.h>
#include <common.h>
#include <config.h>
#include <status.h>
#include <structures.h>
#include <version.h>
#include <openssl/rsa.h>

int	be_daemon = DAEMONMODE ;		/* 0 = Run by inetd */
						/* 2 = Background     */
int	state ;	 					/* state of child */
int	controlsock ;			/* control socket for the server*/
int	msgsock ;			/* control socket for the child */
struct sockaddr_in his_addr;	/* Remote adresse */
struct sockaddr_in ctrl_addr;	/* Local adresse */
int	flagsighup ;
int	killdone= 0 ;
pid_t pid_child[MAXPORT] ;
int	childendinerror ;
char currentfilename[MAXLENFILE];
char currentusername[MAXLEN] ;
int	unlinkfile ;
pid_t	fatherpid ;
RSA		*myrsa ;

void reply(int n, char *str) ;
void clean_child() ;
void exit_clean() ;

main (argc,argv,envp)
	int 	argc ;
	char	**argv ;
	char	**envp ;
{
	int	addrlen ;
	struct timeval	wait_timer;
	fd_set	selectmask ; /* Select mask */
	int	nfds ; /* Max number of file descriptor */
	int	retcode ;
	int	i ;
	char tmpchar[50] ;
	
	sprintf(tmpchar,"bbftpd v%s",VERSION) ;
	openlog(tmpchar, LOG_PID | LOG_NDELAY, LOG_DAEMON);
	
	if ( be_daemon != 0 ) {
		do_daemon(argc, argv, envp);
	} else {
		char	buffrand[NBITSINKEY] ;
		struct timeval tp ;
		unsigned int seed ;
		/*
		** Load the error message from the crypto lib
		*/
		ERR_load_crypto_strings() ;
		/*
		** Initialize the buffrand buffer which is giong to be used to initialize the 
		** random generator
		*/
		/*
		** Take the usec to initialize the random session
		*/
		gettimeofday(&tp,NULL) ;
		seed = tp.tv_usec ;
		srandom(seed) ;
		for (i=0; i < sizeof(buffrand) ; i++) {
			buffrand[i] = random() ;
		}
		/*
		** Initialize the random generator
		*/
		RAND_seed(buffrand,NBITSINKEY) ;
	}
	fatherpid = getpid() ;
	/* Get the remote address */
	addrlen = sizeof(his_addr);
	if (getpeername(msgsock, (struct sockaddr *) &his_addr, &addrlen) < 0) {
		syslog(LOG_ERR, "getpeername (%s): %m", argv[0]);
        exit(1);
	}
	addrlen = sizeof(ctrl_addr);
	if (getsockname(msgsock, (struct sockaddr *) &ctrl_addr, &addrlen) < 0) {
		syslog(LOG_ERR, "getsockname (%s): %m", argv[0]);
		exit(1);
	}
	syslog(LOG_NOTICE,"Getting new bbftp connexion from : %s",inet_ntoa(his_addr.sin_addr)) ;
/*
** Set the socket as non blocking
*/
	if ( fcntl(msgsock,F_SETFL,O_NONBLOCK)  < 0 ) {
		close(msgsock) ;
		syslog(LOG_ERR,"Cannot fcntl O_NONBLOCK on control socket : %m ") ;
		exit(1) ;
	}
	
/*
** Set the signals
*/
	if ( set_signals() < 0 ) {
		exit(1) ;
	}
	for ( i=0 ; i< MAXPORT ; i++) {
		pid_child[i] = 0 ;
	}
	
	/*
	** Send the encryption supported 
	*/
	
	sendcrypt() ;
	
	/* set the state */
	state = S_CONN ;
	unlinkfile = 0 ;

	/* Waiting for something on the stdin */
	
	nfds = sysconf(_SC_OPEN_MAX) ;
	for (;;) {
		/*
		** Initialize the selectmask
		*/
		FD_ZERO(&selectmask) ;
		FD_SET(msgsock,&selectmask) ;
		/*
		** Depending on the state set a timer or not 
		*/
		wait_timer.tv_sec  = 900 ;
		wait_timer.tv_usec = 0 ;
		switch (state) {
			case S_CONN : {
				/*
				** Timer of 10s between connection and MSG_LOG
				** message
				*/
				wait_timer.tv_sec  = 10 ;
				wait_timer.tv_usec = 0 ;
				break ;
			}
			case S_SENDING : 
			case S_RECEIVING : {
				/*
				** No timer while receiving or sending
				*/
				wait_timer.tv_sec  = 0 ;
				wait_timer.tv_usec = 0 ;
				break ;
			}
			default : {
				/*
				** Timer of 900s between commands
				*/
				wait_timer.tv_sec  = 900 ;
				wait_timer.tv_usec = 0 ;
				break ;
			}
		}
		if ( (retcode = select(nfds,&selectmask,0,0,(wait_timer.tv_sec == 0) ? NULL : &wait_timer) ) == -1 ) {
			if ( errno != EINTR ) {
				syslog(LOG_ERR,"Select error : %m") ;
			}
		} else if ( retcode == 0 ) {
			syslog(LOG_NOTICE,"Time OUT ") ;
			clean_child() ;
			exit_clean() ;
			exit(0) ;
		} else {
			/*
			** At this stage we can only receive a command
			*/
			if ( readcontrol() < 0 ) { 
				clean_child() ;
				exit_clean() ;
				exit(0) ;
			} else {
			}
		}
	}
	
	
}

void clean_child() 
{
	int	i ;
	
	if ( killdone == 0 ) {
		killdone = 1 ;
		for ( i=0 ; i<MAXPORT ; i++) {
			if ( pid_child[i] != 0 ) {
				syslog(LOG_NOTICE,"Killing child %d",pid_child[i]) ;
				kill(pid_child[i],SIGKILL) ;
			}
		}
	}
}

void exit_clean() {
	
	switch (state) {
		case S_CONN : {
			return ;
		}
		case S_LOGGED :
		case S_SENDING:
		case S_RECEIVING : {
			syslog(LOG_NOTICE,"User %s disconnected",currentusername) ;
			return ;
		}
		default :{
			return ;
		}
	}
}

#ifndef WORDS_BIGENDIAN
my64_t convertlong(my64_t v) {
	struct bb {
		int	fb ;
		int sb ;
	} ;
	struct bb *bbpt ;
	int 	tmp ;
	my64_t	tmp64 ;
	
	tmp64 = v ;
	bbpt = (struct bb *) &tmp64 ;
	tmp = bbpt->fb ;
	bbpt->fb = ntohl(bbpt->sb) ;
	bbpt->sb = ntohl(tmp) ;	
	return tmp64 ;
}
#endif 
