/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 Create the receive socket and connect to it.

RETURN VALUE : 
 In case of failure no socket is opened and -1 is returned. 
 If connect fail due to interrupted system call 0 is returned.
 In case of success the socket is returned. 
  
 createreceivesock.c v 1.3.0 2000/03/20
                     v 1.4.0 2000/03/20 - Correct bug on sprintf ( forgot the 
					 	                  socknum)
									      Bug shown by gfarrache
					 v 1.6.0 2000/03/24	- Set Send buffer to TCPWINSIZE for
					 					  get option
					 v 1.6.1 2000/03/24	- Portage to OSF1
					 					- Set the socket in non block mode
					 v 1.8.7 2000/05/24	- Change headers

*****************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include <common.h>

extern struct sockaddr_in his_addr;		/* Remote adresse */
extern struct sockaddr_in ctrl_addr;	/* Local adresse */

int createreceivesock(port,socknum,logmessage)
	int		port ;
	int		socknum ;
	char	*logmessage ;
{
	int	sock ;
	struct sockaddr_in data_source ;
	int	on = 1 ;
	int tcpwinsize = TCPWINSIZE ;
	int	addrlen ;
	int	retcode ;

	sock = socket ( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
	if ( sock < 0 ) {
		syslog(LOG_ERR, "Cannot create receive socket %d : %m",socknum);
		sprintf(logmessage,"Cannot create receive socket %d : %s",socknum,strerror(errno)) ;
		return (-1) ;
	}
	if ( setsockopt(sock,SOL_SOCKET, SO_REUSEADDR,(char *)&on,sizeof(on)) < 0 ) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot set SO_REUSEADDR on receive socket %d : %m ",socknum) ;
		sprintf(logmessage,"Cannot set SO_REUSEADDR on receive socket %d : %s",socknum,strerror(errno)) ;
		return (-1) ;
	}
	if ( setsockopt(sock,SOL_SOCKET, SO_RCVBUF,(char *)&tcpwinsize,sizeof(tcpwinsize)) < 0 ) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot set SO_RCVBUF on receive socket %d : %m ",socknum) ;
		sprintf(logmessage,"Cannot set SO_RCVBUF on receive socket %d : %s",socknum,strerror(errno)) ;
		return (-1) ;
	}
	if ( setsockopt(sock,SOL_SOCKET, SO_SNDBUF,(char *)&tcpwinsize,sizeof(tcpwinsize)) < 0 ) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot set SO_SNDBUF on receive socket %d : %m ",socknum) ;
		sprintf(logmessage,"Cannot set SO_SNDBUF on receive socket %d : %s",socknum,strerror(errno)) ;
		return (-1) ;
	}
	if ( setsockopt(sock,IPPROTO_TCP, TCP_NODELAY,(char *)&on,sizeof(on)) < 0 ) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot set TCP_NODELAY on receive socket %d : %m ",socknum) ;
		sprintf(logmessage,"Cannot set TCP_NODELAY on receive socket %d : %s",socknum,strerror(errno)) ;
		return (-1) ;
	}
	data_source.sin_family = AF_INET;
#ifdef SUNOS
	data_source.sin_addr = ctrl_addr.sin_addr;
#else
	data_source.sin_addr.s_addr = INADDR_ANY;
#endif
	data_source.sin_port = htons(ntohs(ctrl_addr.sin_port) - 1);
	if ( bind(sock, (struct sockaddr *) &data_source,sizeof(data_source)) < 0) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot bind on receive socket %d : %m ",socknum) ;
		sprintf(logmessage,"Cannot bind on receive socket %d : %s",socknum,strerror(errno)) ;
		return (-1) ;
	}
	data_source.sin_addr = his_addr.sin_addr;
	data_source.sin_port = htons(port);
	addrlen = sizeof(data_source) ;
	retcode = connect(sock,(struct sockaddr *) &data_source,addrlen) ;
	if ( retcode < 0 ) {
		if ( errno == EINTR ) {
			/*
			** In this case we are going to tell the calling program 
			** to retry so we close the socket and set the return code
			** to zero 
			*/
			close(sock) ;
			syslog(LOG_ERR,"Cannot connect receive socket %d : %m, telling calling to retry ",socknum) ;
			return 0 ;
		} else {
			close(sock) ;
			syslog(LOG_ERR,"Cannot connect receive socket %d : %m ",socknum) ;
			sprintf(logmessage,"Cannot connect receive socket %d : %s",socknum,strerror(errno)) ;
			return (-1) ;
		}
	}
	if ( fcntl(sock,F_SETFL,O_NONBLOCK) < 0 ) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot fcntl O_NONBLOCK on receive socket %d : %m ",socknum) ;
		sprintf(logmessage,"Cannot fcntl O_NONBLOCK on receive socket %d : %s",socknum,strerror(errno)) ;
		return (-1) ;
	}
	return sock ;
}
