/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN:
 		0  login OK
		-1 login failed
 
 loginsequence.c v 1.6.0  2000/03/24	- Creation of the routine. This part of
 									  code was contained in readcontrol.c
				 v 1.8.0  2000/04/14	- Introduce RSA Cryptage
				 v 1.8.3  2000/04/18	- Change the reply message in order
				 						  to stop retry in case of unkown user
										  or incorrect password
 				 v 1.8.7  2000/05/24	- Modify headers
 				 v 1.8.11 2000/05/24	- Add Pam for Linux (and only pam...)
				 v 1.9.0  2000/08/18	- Use configure to help portage
				 						- Change the reply message (for future
										  use)
									  
 *****************************************************************************/

#include <crypt.h>
#include <netinet/in.h>
#include <pwd.h>
#include <syslog.h>
#include <sys/types.h>
#include <unistd.h>

#include <bbftpd.h>
#include <config.h>
#include <common.h>
#include <structures.h>
#include <version.h>

#ifdef SHADOW_PASSWORD
#include <shadow.h>
#endif
#ifdef USE_PAM
#include <security/pam_appl.h>
#endif
#ifdef HAVE_SECURITY_PASS
#include <userpw.h>
#endif

extern int msgsock ;
extern char currentusername[MAXLEN] ;

#ifdef USE_PAM
char tmpchar[50] ;

static char *PAM_password;
/* PAM conversation function
 * We assume that echo on means login name, and echo off means password.
 */
static int PAM_conv(int num_msg, const struct pam_message **msg, struct pam_response **resp, void *appdata_ptr)
{
	int replies = 0;
	struct pam_response *reply = NULL;

#define COPY_STRING(s) (s) ? strdup(s) : NULL

	reply = (struct pam_response *) malloc(sizeof(struct pam_response) * num_msg);
	if (!reply) return PAM_CONV_ERR;

	for (replies = 0; replies < num_msg; replies++) {
		switch (msg[replies]->msg_style) {
		case PAM_PROMPT_ECHO_ON:
			return PAM_CONV_ERR;
    		break;
		case PAM_PROMPT_ECHO_OFF:
    		reply[replies].resp_retcode = PAM_SUCCESS;
    		reply[replies].resp = COPY_STRING(PAM_password);
    		/* PAM frees resp */
    		break;
		case PAM_TEXT_INFO:
    		/* ignore it... */
			reply[replies].resp_retcode = PAM_SUCCESS;
    		reply[replies].resp = NULL;
    		break;
		case PAM_ERROR_MSG:
    		/* ignore it... */
			reply[replies].resp_retcode = PAM_SUCCESS;
    		reply[replies].resp = NULL;
    		break;
		default:
		   /* Must be an error of some sort... */
    		return PAM_CONV_ERR;
		}
	}
	*resp = reply;
	return PAM_SUCCESS;
}

static struct pam_conv PAM_conversation =
{
	&PAM_conv,
	NULL
};
#endif

int loginsequence() {
	
	int		retcode ;
	char	receive_buffer[MAXMESSLEN] ;
	struct mess_sec *msg_sec ;

	char	username[MAXLEN] ;
	char	password[MAXLEN] ;
	struct	passwd	*uspass ;
	char	logmessage[256] ;
#ifdef SHADOW_PASSWORD
	struct	spwd	*sunpass ;
#endif
#ifdef HAVE_SECURITY_PASS
	struct	userpw	*secpass ;
#endif
	char	*calcpass ;
#ifdef AFS
	char	*inst = "";
#endif
#ifdef USE_PAM
	pam_handle_t *pamh=NULL;
#endif
	sprintf(logmessage,"bbftpd version %s",VERSION) ;
	/*
	** Read the crypt type
	*/
	if ( retcode = readmessage(msgsock,receive_buffer,CRYPTMESSLEN,CONTROLSOCKTO) < 0 ) {
		/*
		** Error ...
		*/
		return(retcode) ;
	}
	msg_sec = (struct mess_sec *)receive_buffer ;
	switch (msg_sec->crtype) {
		case CRYPT_RSA_PKCS1_OAEP_PADDING :
			/*
			** RSA
			*/
			if ( retcode = readmessage(msgsock,receive_buffer,RSAMESSLEN,CONTROLSOCKTO) < 0) {
				return(retcode) ;
			}
			if ( decodersapass(receive_buffer,username,password) < 0 ) {
				syslog(LOG_ERR,"Decode error") ;
				strcat(logmessage," : BAD") ;
				reply(MSG_BAD,logmessage) ;
				return -1 ;
			}
			break ;
		default :
			syslog(LOG_ERR,"Unkwown encryption %d",msg_sec->crtype) ;
			strcat(logmessage," : BAD") ;
			reply(MSG_BAD,logmessage) ;
			return -1 ;
	}
	/*
	** Here we check the username and pass and set the default dir
	*/
	if ( (uspass = getpwnam(username)) == NULL ) {
		syslog(LOG_ERR,"Unkwown user %s",username) ;
		strcat(logmessage," : BAD NO RETRY") ;
		reply(MSG_BAD_NO_RETRY,logmessage) ;
		return -1 ;
	}
/*
** If the AFS flag is set it take precedence on all other mechanism
*/
#ifdef AFS
	setpag() ;
	retcode = ka_UserAuthenticate(username,inst,0,password,0,&calcpass) ;
	if ( retcode != 0 ) {
		syslog(LOG_ERR,"ka_UserAuthenticate message : %s ",calcpass) ;
		/*
		** Check local user 
		*/
		if ( (strcmp(uspass->pw_passwd,"x") == 0) || 
			 (strcmp(uspass->pw_passwd,"X") == 0) ) {
#ifdef SHADOW_PASSWORD
			if ( (sunpass = getspnam(username)) == NULL ) {
				syslog(LOG_ERR,"Unkwown user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
			calcpass = (char *) crypt(password,sunpass->sp_pwdp) ;
			if ( strcmp(calcpass,sunpass->sp_pwdp) != 0 ) {
				syslog(LOG_ERR,"Incorrect password user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
#elif defined(HAVE_SECURITY_PASS)
			if ( (secpass = getuserpw(username)) == NULL ) {
				syslog(LOG_ERR,"Unkwown user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
			calcpass = (char *) crypt(password,secpass->upw_passwd) ;
			if ( strcmp(calcpass,secpass->upw_passwd) != 0 ) {
				syslog(LOG_ERR,"Incorrect password user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
#else

			syslog(LOG_ERR,"No Password user %s",username) ;
			strcat(logmessage," : BAD NO RETRY") ;
			reply(MSG_BAD_NO_RETRY,logmessage) ;
			return -1 ;
#endif
		} else {
			calcpass = (char *) crypt(password,uspass->pw_passwd) ;
			if ( strcmp(calcpass,uspass->pw_passwd) != 0 ) {
				syslog(LOG_ERR,"Incorrect password user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
		}
	}
#elif defined(USE_PAM)
	/*
	** For PAM
	*/
	strcat(logmessage," : BAD NO RETRY") ;
	PAM_password = password ;
	retcode = pam_start("bbftp", username, &PAM_conversation, &pamh);
#define PAM_BAIL if (retcode != PAM_SUCCESS) { pam_end(pamh, 0); syslog(LOG_ERR,"PAM error (%d) user %s",retcode,username) ;reply(MSG_BAD_NO_RETRY,logmessage) ;return -1; }
	PAM_BAIL;
	retcode = pam_authenticate(pamh, PAM_SILENT);
	PAM_BAIL;
	retcode = pam_acct_mgmt(pamh, 0);
	PAM_BAIL;
	pam_end(pamh, PAM_SUCCESS);
	/*
	** Reopen the lag as the pam functions close it
	*/
	sprintf(tmpchar,"bbftpd v%s",VERSION) ;
	openlog(tmpchar, LOG_PID | LOG_NDELAY, LOG_DAEMON);
#else
	/*
	** For the others
	*/
	if ( (strcmp(uspass->pw_passwd,"x") == 0) || 
		 (strcmp(uspass->pw_passwd,"X") == 0)) {
#ifdef SHADOW_PASSWORD
			if ( (sunpass = getspnam(username)) == NULL ) {
				syslog(LOG_ERR,"Unkwown user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
			calcpass = (char *) crypt(password,sunpass->sp_pwdp) ;
			if ( strcmp(calcpass,sunpass->sp_pwdp) != 0 ) {
				syslog(LOG_ERR,"Incorrect password user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
#elif defined(HAVE_SECURITY_PASS)
			if ( (secpass = getuserpw(username)) == NULL ) {
				syslog(LOG_ERR,"Unkwown user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
			calcpass = (char *) crypt(password,secpass->upw_passwd) ;
			if ( strcmp(calcpass,secpass->upw_passwd) != 0 ) {
				syslog(LOG_ERR,"Incorrect password user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
#else
			syslog(LOG_ERR,"No Password user %s",username) ;
			strcat(logmessage," : BAD NO RETRY") ;
			reply(MSG_BAD_NO_RETRY,logmessage) ;
			return -1 ;
#endif
		} else {
			calcpass = (char *) crypt(password,uspass->pw_passwd) ;
			if ( strcmp(calcpass,uspass->pw_passwd) != 0 ) {
				syslog(LOG_ERR,"Incorrect password user %s",username) ;
				strcat(logmessage," : BAD NO RETRY") ;
				reply(MSG_BAD_NO_RETRY,logmessage) ;
				return -1 ;
			}
		}
#endif
	/*
	** Set the uid and gid of the process
	*/
	if ( setgid(uspass->pw_gid) < 0 ) {
		syslog(LOG_ERR,"Error setgid user %s : %m",username) ;
		strcat(logmessage," : BAD") ;
		reply(MSG_BAD,logmessage) ;
		return -1 ;
	}
	if ( setuid(uspass->pw_uid) < 0 ) {
		syslog(LOG_ERR,"Error setuid user %s : %m",username) ;
		strcat(logmessage," : BAD") ;
		reply(MSG_BAD,logmessage) ;
		return -1 ;
	}
	if ( uspass->pw_dir == NULL ) {
		syslog(LOG_ERR,"No home directory user %s : %m",username) ;
		strcat(logmessage," : No home directory") ;
		reply(MSG_BAD,logmessage) ;
		return -1 ;
	}
	if ( chdir(uspass->pw_dir) < 0) {
		syslog(LOG_ERR,"Error chdir user %s : %m",username) ;
		strcat(logmessage," : Unable to change directory") ;
		reply(MSG_BAD,logmessage) ;
		return -1 ;
	}
				
	syslog(LOG_NOTICE,"User %s connected",username) ;
	strcpy(currentusername,username) ;
	return 0 ;
}
