/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 set_signals.c v 0.0.0  1999/11/24
 			   v 1.6.0  2000/03/25	- Make the unlink conditionnal (not
			   						  to be done in the case of a get)
  			   v 1.8.3  2000/04/18	- Implement better return code and
			   						  MSG_BAD_NO_RETRY
  			   v 1.8.10 2000/04/18	- Suppress SIGINFO as it is not used
									  (Mainly for Linux)
			   v 1.9.0  2000/08/18	- Use configure to help portage

*****************************************************************************/
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <syslog.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <bbftpd.h>
#include <daemon.h>
#include <status.h>
#include <structures.h>

extern int pid_child[MAXPORT] ;
extern int flagsighup;
extern int childendinerror ;
extern int state ;
extern char currentfilename[MAXLENFILE];
extern int killdone ;
extern int unlinkfile ;
extern pid_t	fatherpid ;

void clean_child() ; 

int checkendchild(status)
	int status ;
{
	
	if (WEXITSTATUS(status) == 0 ) {
		if ( WIFSIGNALED(status) == 0 ) {
			return(0) ;
		} else {
			/*
			** In order to differenciate signal and ernno we choose
			** to return ENOEXEC in case of signal
			*/
			return(ENOEXEC) ;
		}
	} else {
		return(WEXITSTATUS(status)) ;
	}			 
}

void in_sigchld(sig) 
int		sig ;
{
	int	pid ;
	int	status ;
	int	totpid ;
	int i ;
	char logmessage[256] ;
	int	retcode ;
	
	/*
	** We are going to check if the process generating this
	** signal was not already detected as dead. This was causing 
	** a send of a MSG_OK if all process have been dectected dead
	** througn the waitpid routine and the signal was pending.
	** If all processes have been detected dead before childendinerror 
	** to 0, totpid also so we send a reply MSS_OK
	*/
	totpid = 0 ;
	for ( i=0 ; i< MAXPORT; i++) {
		totpid = totpid + pid_child[i] ;
	}
	if ( totpid == 0 ) {
		/*
		** All Ended
		*/
		return ;
	}
	totpid = 0 ;
/*
** Check the status
*/
	for ( i=0 ; i< MAXPORT; i++) {
		if ( pid_child[i] != 0 ) {
			pid = waitpid(pid_child[i],&status,WNOHANG) ;
			if ( pid == pid_child[i] ) {
				/*
				** the process has ended
				*/
				if ( (retcode = checkendchild(status)) != 0 ) {
					if ( childendinerror == 0 ) {
						childendinerror = 1 ;
						syslog(LOG_ERR,"Child pid %d ends in error status %d",pid,retcode) ;
						if ( unlinkfile == 1 ) unlink(currentfilename) ;
						if (retcode == 255) {
							sprintf(logmessage,"Disk quota excedeed or No Space left on device") ;
							reply(MSG_BAD_NO_RETRY,logmessage) ;
						} else {
							if (retcode ==  ENOEXEC ) {
								sprintf(logmessage,"Interrupted by signal") ;
							} else {
								sprintf(logmessage,"%s",strerror(retcode)) ;
							}
							reply(MSG_BAD,logmessage) ;
						}
					} else {
						syslog(LOG_ERR,"Child pid %d ends in error",pid) ;
					}
				}
				pid_child[i] = 0 ;
			}	
		}
	}
	for ( i=0 ; i< MAXPORT; i++) {
		totpid = totpid + pid_child[i] ;
	}
	if ( totpid == 0 && childendinerror == 0 ) {
		reply(MSG_OK,"OK") ;
		state = S_LOGGED ;
	} 
	if ( childendinerror == 1 ) {
		clean_child() ;
	}
	if (totpid == 0 ) {
		state = S_LOGGED ;
		unlinkfile = 0 ;
		killdone = 0 ;
		childendinerror = 0 ;
	}
	
}
void in_sighup(sig) 
int		sig ;
{
	/* Receive sighup in order to restart*/
	/*
	syslog(LOG_NOTICE,"Receive SIGHUP") ;
	*/
	flagsighup = 1 ;
}

void in_sigterm(sig) 
int		sig ;
{
	if ( fatherpid == getpid() ) {
		/*
		** We are in father
		*/
		/*
		** Unlink the file if we are getting a file
		*/ 
		if ( unlinkfile == 1 ) unlink(currentfilename) ;
		clean_child() ;
		exit() ;
	} else {
		exit(EINTR) ;
	}
}


int set_signals() {
	struct	sigaction	sga ;
	
	sga.sa_handler = SIG_IGN ;
	sigemptyset(&(sga.sa_mask));
	sga.sa_flags = 0   ;
	if ( sigaction(SIGPIPE,&sga,0) < 0 ) {
		syslog(LOG_ERR,"Error sigaction SIGPIPE : %m ") ;
		return(-1) ;
	}
	sga.sa_handler = in_sigchld ;
	sigemptyset(&(sga.sa_mask));
	sga.sa_flags = 0  ;
	if ( sigaction(SIGCHLD,&sga,0) < 0 ) {
		syslog(LOG_ERR,"Error sigaction SIGCHLD : %m ") ;
		return(-1) ;
	}
	sga.sa_handler = in_sighup ;
	sigemptyset(&(sga.sa_mask));
	sga.sa_flags = 0  ;
	if ( sigaction(SIGHUP,&sga,0) < 0 ) {
		syslog(LOG_ERR,"Error sigaction SIGHUP : %m ") ;
		return(-1) ;
	}
	sga.sa_handler = in_sigterm ;
	sigemptyset(&(sga.sa_mask));
	sga.sa_flags = 0  ;
	if ( sigaction(SIGTERM,&sga,0) < 0 ) {
		syslog(LOG_ERR,"Error sigaction SIGTERM : %m ") ;
		return(-1) ;
	}
	return 0 ;
}
#ifdef SUNOS
int sigsetmask(mask)
        int mask;
{
        return(sigprocmask(SIG_SETMASK,(sigset_t *)&mask,(sigset_t *)NULL));
}
struct sigvec
{ void (*sv_handler)();
  int sv_mask;
  int sv_flags;
};

int sigvec(signal,invec,outvec)
        int signal;
        struct sigvec *invec,*outvec;
{
        struct sigaction action,oaction;
        int rc;

        action.sa_handler=invec->sv_handler;
        (void) memset(&(action.sa_mask),0,sizeof(sigset_t));
        (void) memmove(&(action.sa_mask),&(invec->sv_mask),sizeof(int));
        action.sa_flags=invec->sv_flags;

        rc=sigaction(signal,&action,&oaction);
        if (rc) return(rc);

        outvec->sv_handler=oaction.sa_handler;
        (void) memmove(&(outvec->sv_mask),&(oaction.sa_mask),sizeof(int));
        outvec->sv_flags=oaction.sa_flags;

        return(0);
}
#endif
